# pylint: disable-msg=too-many-lines
"""Common code for P-Roc and P3-Roc."""
import abc
import asyncio
import logging
from logging import Logger

import sys
from threading import Thread

import time
from typing import List, Union, Tuple, Optional, Set

from mpf._version import log_url
from mpf.core.utility_functions import Util
from mpf.core.platform_batch_light_system import PlatformBatchLightSystem
from mpf.platforms.interfaces.servo_platform_interface import ServoPlatformInterface

from mpf.platforms.p_roc_devices import PROCSwitch, PROCMatrixLight, PDBLED, PDBLight, PDBCoil, PDBSwitch, PdLedServo, \
    PdLedStepper

from mpf.platforms.interfaces.light_platform_interface import LightPlatformInterface
from mpf.core.platform import SwitchPlatform, DriverPlatform, LightsPlatform, SwitchSettings, DriverSettings, \
    SwitchConfig, ServoPlatform, StepperPlatform, RepulseSettings

from mpf.exceptions.runtime_error import MpfRuntimeError


# pylint: disable-msg=ungrouped-imports
try:    # pragma: no cover
    import pinproc
    PINPROC_IMPORTED = True
    IMPORT_ERROR = None
except ImportError:     # pragma: no cover
    try:
        if sys.platform == 'darwin':
            from mpf.platforms.pinproc.osx import pinproc
        elif sys.platform == 'win32':
            from mpf.platforms.pinproc.windows import pinproc
        else:
            raise ImportError

        PINPROC_IMPORTED = True
        IMPORT_ERROR = None

    except ImportError as e:
        PINPROC_IMPORTED = False
        pinproc = None
        IMPORT_ERROR = e


class ProcProcess:

    """External pinproc process."""

    def __init__(self):
        """Initialise process."""
        self.proc = None
        self.dmd = None
        self.loop = None
        self.stop_future = None
        self.trace = None
        self.log = None

    def start_pinproc(self, machine_type, loop, trace, log):
        """Initialise libpinproc."""
        self.loop = loop
        asyncio.set_event_loop(loop)
        self.stop_future = asyncio.Future()
        self.trace = trace
        assert log is not None
        self.log = log  # type: Logger
        while not self.proc:
            try:
                self.proc = pinproc.PinPROC(machine_type)
            except IOError as e:     # pragma: no cover
                self.log.warning("Failed to instantiate pinproc.PinPROC(%s): %s. Is your P/P3-Roc connected "
                                 "and powered up?", machine_type, e)
                self.log.info("Will retry creating PinPROC in 1s.")
                time.sleep(1)
                continue

            try:
                self.proc.reset(1)
            except IOError as e:  # pragma: no cover
                self.log.warning("Failed to reset P/P3-Roc: %s. Is your P/P3-Roc connected and powered up?", e)
                self.log.info("Will retry creating PinPROC and resetting it in 1s.")
                time.sleep(1)
                continue

    def start_proc_process(self, machine_type, loop, trace, log):
        """Run the pinproc communication."""
        asyncio.set_event_loop(loop)
        self.start_pinproc(machine_type, loop, trace, log)

        loop.run_until_complete(self.stop_future)
        loop.close()

    def stop(self):
        """Stop thread."""
        self.stop_future.set_result(True)

    @staticmethod
    def _sync(num):
        return "sync", num

    def _write_data_batch(self, data):
        for cmd in data:
            self.proc.write_data(*cmd)
        self.proc.flush()
        if self.trace:
            self.log.debug("pinproc.PinPROC.write_data(%s)", data)
            self.log.debug("pinproc.PinPROC.flush()")

    async def run_command(self, cmd, *args):
        """Run command in proc thread."""
        try:
            if cmd.startswith("_"):
                return getattr(self, cmd)(*args)

            if self.trace:
                assert self.log is not None
                result = getattr(self.proc, cmd)(*args)
                self.log.debug("pinproc.PinPROC.%s%s -> %s", cmd, args, result)
                return result

            return getattr(self.proc, cmd)(*args)
        except IOError as error:  # pragma: no cover
            raise MpfRuntimeError("Communication with P/P3-Roc broke down. Check USB cable and power supply.", 2,
                                  self.log.name) from error

    def _dmd_send(self, data):
        if not self.dmd:
            # size is hardcoded here since 128x32 is all the P-ROC hw supports
            self.dmd = pinproc.DMDBuffer(128, 32)

        self.dmd.set_data(data)
        self.proc.dmd_draw(self.dmd)

    async def read_events_and_watchdog(self, poll_sleep):
        """Return all events and tickle watchdog."""
        try:
            while not self.stop_future.done():
                events = self.proc.get_events()
                self.proc.watchdog_tickle()
                self.proc.flush()
                if events:
                    return list(events)

                await asyncio.sleep(poll_sleep)

            return []
        except IOError as error:  # pragma: no cover
            raise MpfRuntimeError("Communication with P/P3-Roc broke down. Check USB cable and power supply.", 2,
                                  self.log.name) from error


# pylint does not understand that this class is abstract
# pylint: disable-msg=abstract-method
# pylint: disable-msg=too-many-instance-attributes
class PROCBasePlatform(LightsPlatform, SwitchPlatform, DriverPlatform, ServoPlatform, StepperPlatform,
                       metaclass=abc.ABCMeta):

    """Platform class for the P-Roc and P3-ROC hardware controller.

    Args:
    ----
        machine: The MachineController instance.
    """

    __slots__ = ["pdbconfig", "pinproc", "proc", "log", "hw_switch_rules", "version", "revision", "hardware_version",
                 "dipswitches", "machine_type", "event_task", "_late_init_futures",
                 "proc_thread", "proc_process", "proc_process_instance", "_commands_running", "config", "_light_system"]

    def __init__(self, machine):
        """Make sure pinproc was loaded."""
        super().__init__(machine)

        if not PINPROC_IMPORTED:
            raise MpfRuntimeError('Could not import "pinproc". Either the library is not installed or is missing '
                                  'some of its dependencies. Check the install instructions for your OS in '
                                  'Multimorphic section of the MPF docs. You can run mpf with "-X" to use virtual '
                                  'hardware in the meantime.', 3, 'P-Roc') from IMPORT_ERROR

        self.pdbconfig = None
        self.pinproc = pinproc
        self.log = None
        self.hw_switch_rules = {}
        self.version = None
        self.revision = None
        self.hardware_version = None
        self.dipswitches = None
        self.event_task = None
        self.proc_thread = None
        self.proc_process = None
        self.proc_process_instance = None
        self._commands_running = 0
        self.config = {}
        self._light_system = None
        self.machine_type = None
        self._late_init_futures = []

    def _decrement_running_commands(self, future):
        del future
        self._commands_running -= 1

    def run_proc_cmd(self, cmd, *args):
        """Run a command in the p-roc thread and return a future."""
        if self.debug:
            self.debug_log("Calling P-Roc cmd: %s (%s)", cmd, args)
        future = asyncio.wrap_future(
            asyncio.run_coroutine_threadsafe(self.proc_process.run_command(cmd, *args), self.proc_process_instance))
        future.add_done_callback(Util.raise_exceptions)
        return future

    def run_proc_cmd_no_wait(self, cmd, *args):
        """Run a command in the p-roc thread."""
        if self.debug:
            self.debug_log("Calling P-Roc cmd (no wait): %s (%s)", cmd, args)
        asyncio.run_coroutine_threadsafe(self.proc_process.run_command(cmd, *args), self.proc_process_instance)

    def run_proc_cmd_sync(self, cmd, *args):
        """Run a command in the p-roc thread and return the result."""
        return self.machine.clock.loop.run_until_complete(self.run_proc_cmd(cmd, *args))

    async def initialize(self):
        """Set machine vars."""
        await super().initialize()
        await self.connect()
        self.machine.variables.set_machine_var("p_roc_version", self.version)
        '''machine_var: p_roc_version

        desc: Holds the firmware version number of the P-ROC or P3-ROC controller that's
        attached to MPF.
        '''

        self.machine.variables.set_machine_var("p_roc_revision", self.revision)
        '''machine_var: p_roc_revision

        desc: Holds the firmware revision number of the P-ROC or P3-ROC controller
        that's attached to MPF.
        '''

        self.machine.variables.set_machine_var("p_roc_hardware_version", self.hardware_version)
        '''machine_var: p_roc_hardware_version

        desc: Holds the hardware version number of the P-ROC or P3-ROC controller
        that's attached to MPF.
        '''

        self._light_system = PlatformBatchLightSystem(self.machine.clock, self._send_multiple_light_update,
                                                      self.machine.config['mpf']['default_light_hw_update_hz'],
                                                      65535)

    async def _send_multiple_light_update(self, sequential_brightness_list: List[Tuple[PDBLED, float, int]]):
        """Update a list of light at once."""
        first_light, _, common_fade_ms = sequential_brightness_list[0]
        board = first_light.board
        command_buffer = []     # type: List[Tuple[int, int, int]]
        if common_fade_ms > 0:
            # set the fade time
            self._write_fade_time_buffered(board, int(common_fade_ms / 4), command_buffer)

        # set address of first board
        self._write_addr_buffered(board, first_light.address, command_buffer)

        if self.debug:
            self.debug_log("Fading %s lights with %s fade_ms ",
                           len(sequential_brightness_list), common_fade_ms)

        for light, brightness, fade_ms in sequential_brightness_list:
            if light.polarity:
                value = 255 - int(brightness * 255)
            else:
                value = int(brightness * 255)

            if self.debug:
                self.debug_log("Setting color %s with fade_ms %s to %s-%s",
                               value, fade_ms, light.board, light.address)
            if common_fade_ms > 0:
                self._write_fade_color_buffered(board, value, command_buffer)
            else:
                self._write_color_buffered(board, value, command_buffer)

        self.run_proc_cmd_no_wait("_write_data_batch", command_buffer)

    async def start(self):
        """Start listening for switches."""
        if self._late_init_futures:
            await asyncio.wait(self._late_init_futures)

        self.event_task = self.machine.clock.loop.create_task(self._poll_events())
        self.event_task.add_done_callback(Util.raise_exceptions)
        self._light_system.start()

    def process_events(self, events):
        """Process events from the P-Roc."""
        raise NotImplementedError()

    async def _poll_events(self):
        poll_sleep = 1 / self.machine.config['mpf']['default_platform_hz']
        while True:
            events = await asyncio.wrap_future(
                asyncio.run_coroutine_threadsafe(self.proc_process.read_events_and_watchdog(poll_sleep),
                                                 self.proc_process_instance))
            if events:
                self.process_events(events)

            await asyncio.sleep(poll_sleep)

    def stop(self):
        """Stop proc."""
        if self._light_system:
            self._light_system.stop()
        if self.proc_process and self.proc_process_instance:
            self.proc_process_instance.call_soon_threadsafe(self.proc_process.stop)
        if self.proc_thread:
            self.debug_log("Waiting for pinproc thread.")
            self.proc_thread.join()
            self.proc_thread = None
            self.debug_log("pinproc thread finished.")

    def _start_proc_process(self):
        self.proc_process = ProcProcess()
        if self.config["use_separate_thread"]:
            # Create a new loop
            self.proc_process_instance = asyncio.new_event_loop()
            # Assign the loop to another thread
            self.proc_thread = Thread(target=self.proc_process.start_proc_process,
                                      args=(self.machine_type, self.proc_process_instance,
                                            self.config['trace_bus'] and self.config['debug'],
                                            self.log))
            self.proc_thread.start()

        else:
            # use existing loop
            self.proc_process_instance = self.machine.clock.loop
            self.proc_process.start_pinproc(loop=self.machine.clock.loop, machine_type=self.machine_type,
                                            trace=self.config['trace_bus'] and self.config['debug'],
                                            log=self.log)

    async def connect(self):
        """Connect to the P-ROC.

        Keep trying if it doesn't work the first time.
        """
        self.log.info("Connecting to P-ROC")

        self._start_proc_process()

        version_revision = await self.run_proc_cmd("read_data", 0x00, 0x01)

        self.revision = version_revision & 0xFFFF
        self.version = (version_revision & 0xFFFF0000) >> 16
        dipswitches = await self.run_proc_cmd("read_data", 0x00, 0x03)
        self.hardware_version = (dipswitches & 0xF00) >> 8
        self.dipswitches = ~dipswitches & 0x3F

        self.log.info("Successfully connected to P-ROC/P3-ROC. Firmware Version: %s. Firmware Revision: %s. "
                      "Hardware Board ID: %s",
                      self.version, self.revision, self.hardware_version)

        if self.version < 2 or (self.version == 2 and self.revision < 14):
            self.log.warning("Consider upgrading the firmware of your P/P3-Roc to at least 2.14. "
                             "Your version contains known bugs. See: %s",
                             log_url.format("{}-{}-{}".format("RE", self.log.name, 1)))

        # for unknown reasons we have to postpone this a bit after init
        self.machine.delay.add(100, self._configure_pd_led)

    def _configure_pd_led(self):
        """Configure PD-LEDs."""
        for pd_number, config in self.config['pd_led_boards'].items():
            self._write_ws2811_ctrl(pd_number, config['ws281x_low_bit_time'], config['ws281x_high_bit_time'],
                                    config['ws281x_end_bit_time'], config['ws281x_reset_bit_time'])
            self._write_ws2811_range(pd_number, 0, config['ws281x_0_first_address'], config['ws281x_0_last_address'])
            self._write_ws2811_range(pd_number, 1, config['ws281x_1_first_address'], config['ws281x_1_last_address'])
            self._write_ws2811_range(pd_number, 2, config['ws281x_2_first_address'], config['ws281x_2_last_address'])
            self._write_lpd8806_range(pd_number, 0, config['lpd880x_0_first_address'], config['lpd880x_0_last_address'])
            self._write_lpd8806_range(pd_number, 1, config['lpd880x_1_first_address'], config['lpd880x_1_last_address'])
            self._write_lpd8806_range(pd_number, 2, config['lpd880x_2_first_address'], config['lpd880x_2_last_address'])
            self._write_pdled_serial_control(pd_number, (config['use_ws281x_0'] * 1 << 0) +
                                             (config['use_ws281x_1'] * 1 << 1) +
                                             (config['use_ws281x_2'] * 1 << 2) +
                                             (config['use_lpd880x_0'] * 1 << 3) +
                                             (config['use_lpd880x_1'] * 1 << 4) +
                                             (config['use_lpd880x_2'] * 1 << 5) +
                                             (config['use_stepper_0'] * 1 << 8) +
                                             (config['use_stepper_1'] * 1 << 9))

            # configure servos
            self.write_pdled_config_reg(pd_number, 20, (config['use_servo_0'] * 1 << 0) +
                                        (config['use_servo_1'] * 1 << 1) +
                                        (config['use_servo_2'] * 1 << 2) +
                                        (config['use_servo_3'] * 1 << 3) +
                                        (config['use_servo_4'] * 1 << 4) +
                                        (config['use_servo_5'] * 1 << 5) +
                                        (config['use_servo_6'] * 1 << 6) +
                                        (config['use_servo_7'] * 1 << 7) +
                                        (config['use_servo_8'] * 1 << 8) +
                                        (config['use_servo_9'] * 1 << 9) +
                                        (config['use_servo_10'] * 1 << 10) +
                                        (config['use_servo_11'] * 1 << 11))
            self.write_pdled_config_reg(pd_number, 21, config['max_servo_value'])

            # configure steppers
            if config['use_stepper_0'] or config['use_stepper_1']:
                self.write_pdled_config_reg(pd_number, 22, config['stepper_speed'])

    def write_pdled_config_reg(self, board_addr, addr, reg_data):
        """Write a pdled config register.

        Args:
        ----
            board_addr: Address of the board
            addr: Register address
            reg_data: Register data

        Write the 'regData' into the PD-LEDs address register because when writing a config register
        The data (16 bits) goes into the address field, and the address (8-bits) goes into the data field.
        """
        self._write_addr(board_addr, reg_data)
        self._write_reg_data(board_addr, addr)

    def write_pdled_color(self, board_addr, addr, color):
        """Set a color instantly.

        This command will internally increment the index on the PD-LED board.
        Therefore, it will be much more efficient if you set colors for addresses sequentially.
        """
        command_buffer = []
        self._write_addr_buffered(board_addr, addr, command_buffer)
        self._write_color_buffered(board_addr, color, command_buffer)
        self.run_proc_cmd_no_wait("_write_data_batch", command_buffer)

    def _write_addr(self, board_addr, addr):
        """Write an address to pdled."""
        base_reg_addr = 0x01000000 | (board_addr & 0x3F) << 16
        proc_output_module = 3
        proc_pdb_bus_addr = 0xC00

        # Write the low address bits into reg addr 0
        data = base_reg_addr | (addr & 0xFF)
        self.run_proc_cmd_no_wait("write_data", proc_output_module, proc_pdb_bus_addr, data)

        # Write the high address bits into reg addr 6
        data = base_reg_addr | (6 << 8) | ((addr >> 8) & 0xFF)
        self.run_proc_cmd_no_wait("write_data", proc_output_module, proc_pdb_bus_addr, data)

    def _write_reg_data(self, board_addr, data):
        """Write data to pdled."""
        base_reg_addr = 0x01000000 | (board_addr & 0x3F) << 16
        proc_output_module = 3
        proc_pdb_bus_addr = 0xC00

        # Write 0 into reg addr 7, which is the data word, which is actually the address when
        # writing a config write.  The config register is mapped to 0.
        word = base_reg_addr | (7 << 8) | data
        self.run_proc_cmd_no_wait("write_data", proc_output_module, proc_pdb_bus_addr, word)

    @staticmethod
    def _write_fade_time_buffered(board_addr, fade_time, buffer):
        """Set the fade time for a board."""
        proc_output_module = 3
        proc_pdb_bus_addr = 0xC00
        base_reg_addr = 0x01000000 | (board_addr & 0x3F) << 16
        data = base_reg_addr | (3 << 8) | (fade_time & 0xFF)
        buffer.append((proc_output_module, proc_pdb_bus_addr, data))
        data = base_reg_addr | (4 << 8) | ((fade_time >> 8) & 0xFF)
        buffer.append((proc_output_module, proc_pdb_bus_addr, data))

    @staticmethod
    def _write_addr_buffered(board_addr, addr, buffer):
        """Write an address to pdled."""
        base_reg_addr = 0x01000000 | (board_addr & 0x3F) << 16
        proc_output_module = 3
        proc_pdb_bus_addr = 0xC00

        # Write the low address bits into reg addr 0
        data = base_reg_addr | (addr & 0xFF)
        buffer.append((proc_output_module, proc_pdb_bus_addr, data))

        # Write the high address bits into reg addr 6
        data = base_reg_addr | (6 << 8) | ((addr >> 8) & 0xFF)
        buffer.append((proc_output_module, proc_pdb_bus_addr, data))

    @staticmethod
    def _write_color_buffered(board_addr, color, buffer):
        base_reg_addr = 0x01000000 | (board_addr & 0x3F) << 16
        proc_output_module = 3
        proc_pdb_bus_addr = 0xC00

        data = base_reg_addr | (1 << 8) | (color & 0xFF)
        buffer.append((proc_output_module, proc_pdb_bus_addr, data))

    @staticmethod
    def _write_fade_color_buffered(board_addr, fade_color, buffer):
        """Fade the LED at the current index for a board to a certain color."""
        proc_output_module = 3
        proc_pdb_bus_addr = 0xC00
        base_reg_addr = 0x01000000 | (board_addr & 0x3F) << 16
        data = base_reg_addr | (2 << 8) | (fade_color & 0xFF)
        buffer.append((proc_output_module, proc_pdb_bus_addr, data))

    # pylint: disable-msg=too-many-arguments
    def _write_ws2811_ctrl(self, board_addr, lbt, hbt, ebt, rbt):
        self.write_pdled_config_reg(board_addr, 4, lbt)
        self.write_pdled_config_reg(board_addr, 5, hbt)
        self.write_pdled_config_reg(board_addr, 6, ebt)
        self.write_pdled_config_reg(board_addr, 7, rbt)

    def _write_pdled_serial_control(self, board_addr, index_mask):
        self.write_pdled_config_reg(board_addr, 0, index_mask)

    def _write_ws2811_range(self, board_addr, index, first_addr, last_addr):
        self.write_pdled_config_reg(board_addr, 8 + index * 2, first_addr)
        self.write_pdled_config_reg(board_addr, 9 + index * 2, last_addr)

    def _write_lpd8806_range(self, board_addr, index, first_addr, last_addr):
        self.write_pdled_config_reg(board_addr, 16 + index * 2, first_addr)
        self.write_pdled_config_reg(board_addr, 17 + index * 2, last_addr)

    @classmethod
    def _get_event_type(cls, sw_activity, debounced):
        if sw_activity == 0 and debounced:
            return "open_debounced"
        if sw_activity == 0 and not debounced:
            return "open_nondebounced"
        if sw_activity == 1 and debounced:
            return "closed_debounced"
        # if sw_activity == 1 and not debounced:
        return "closed_nondebounced"

    def _add_hw_rule(self, switch: SwitchSettings, coil: DriverSettings, rule, invert=False):
        rule_type = self._get_event_type(switch.invert == invert, switch.debounce)

        if not switch.hw_switch.has_rules:
            raise AssertionError("Switch {} does not support hardware rules.".format(switch.hw_switch))

        if not coil.hw_driver.has_rules:
            raise AssertionError("Driver {} does not support hardware rules.".format(coil.hw_driver))

        # overwrite rules for the same switch and coil combination
        for rule_num, rule_obj in enumerate(switch.hw_switch.hw_rules[rule_type]):
            if rule_obj[0] == switch.hw_switch.number and rule_obj[1] == coil.hw_driver.number:
                del switch.hw_switch.hw_rules[rule_type][rule_num]

        switch.hw_switch.hw_rules[rule_type].append(
            (switch.hw_switch.number, coil.hw_driver.number, rule)
        )

    def _add_pulse_rule_to_switch(self, switch, coil):
        """Add a rule to pulse a coil on switch hit for a certain duration and optional with PWM."""
        # make sure we never set 0 (due to a bug elsewhere) as this would turn the driver on permanently
        assert coil.pulse_settings.duration != 0

        if coil.pulse_settings.power < 1.0:
            pwm_on, pwm_off = coil.hw_driver.get_pwm_on_off_ms(coil.pulse_settings)
            if self.version < 2 or (self.version == 2 and self.revision < 14):
                raise MpfRuntimeError("Your P/P3-Roc firmware contains a known bug with pulsed_patter hardware rules. "
                                      "Please upgrade the firmware to at least 2.14. "
                                      "As a workaround you might remove pulse_power from "
                                      "coil: {}.".format(coil.hw_driver.number),
                                      1, self.log.name)

            self._add_hw_rule(switch, coil,
                              self.pinproc.driver_state_pulsed_patter(coil.hw_driver.state(), pwm_on, pwm_off,
                                                                      coil.pulse_settings.duration, True))
        else:
            self._add_hw_rule(switch, coil,
                              self.pinproc.driver_state_pulse(coil.hw_driver.state(), coil.pulse_settings.duration))

    def _add_pulse_and_hold_rule_to_switch(self, switch: SwitchSettings, coil: DriverSettings):
        """Add a rule to pulse a coil on switch hit for a certain duration and enable the coil with optional PWM.

        The initial pulse will always be at full power and this method will error out if it is set differently.
        """
        # make sure we never set 0 (due to a bug elsewhere) as this would turn the driver on permanently
        assert coil.pulse_settings.duration != 0

        if coil.pulse_settings.power < 1.0:
            self.raise_config_error("Any rules with hold need to have pulse_power set to 1.0. This is a limitation "
                                    "with the P/P3-Roc.", 6)

        if coil.hold_settings.power < 1.0:
            pwm_on, pwm_off = coil.hw_driver.get_pwm_on_off_ms(coil.hold_settings)
            self._add_hw_rule(switch, coil,
                              self.pinproc.driver_state_patter(
                                  coil.hw_driver.state(), pwm_on, pwm_off, coil.pulse_settings.duration, True))
        else:
            # This method is called in the p-roc thread. we can call hw_driver.state()
            self._add_hw_rule(switch, coil, self.pinproc.driver_state_pulse(coil.hw_driver.state(), 0))

    def _add_release_disable_rule_to_switch(self, switch: SwitchSettings, coil: DriverSettings):
        self._add_hw_rule(switch, coil,
                          self.pinproc.driver_state_disable(coil.hw_driver.state()), invert=True)

    def _add_disable_rule_to_switch(self, switch: SwitchSettings, coil: DriverSettings):
        self._add_hw_rule(switch, coil,
                          self.pinproc.driver_state_disable(coil.hw_driver.state()))

    def _add_hold_rule_to_switch(self, switch: SwitchSettings, coil: DriverSettings):
        if coil.hold_settings.power < 1.0:
            pwm_on, pwm_off = coil.hw_driver.get_pwm_on_off_ms(coil.hold_settings)
            self._add_hw_rule(switch, coil,
                              self.pinproc.driver_state_patter(
                                  coil.hw_driver.state(), pwm_on, pwm_off, 0, True))

    def _write_rules_to_switch(self, switch, coil, drive_now):
        for event_type, driver_rules in switch.hw_switch.hw_rules.items():
            driver = []
            for x in driver_rules:
                driver.append(x[2])
            rule = {'notifyHost': bool(switch.hw_switch.notify_on_nondebounce) == event_type.endswith("nondebounced"),
                    'reloadActive': bool(coil.recycle)}
            if drive_now is None:
                self.run_proc_cmd_no_wait("switch_update_rule", switch.hw_switch.number, event_type, rule, driver)
            else:
                self.run_proc_cmd_no_wait("switch_update_rule", switch.hw_switch.number, event_type, rule, driver,
                                          drive_now)

    def set_pulse_on_hit_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit rule on driver."""
        self.debug_log("Setting HW Rule on pulse on hit. Switch: %s, Driver: %s",
                       enable_switch.hw_switch.number, coil.hw_driver.number)
        self._add_pulse_rule_to_switch(enable_switch, coil)
        self._write_rules_to_switch(enable_switch, coil, False)

    def set_pulse_on_hit_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit and release rule to driver."""
        self.debug_log("Setting HW Rule on pulse on hit and relesae. Switch: %s, Driver: %s",
                       enable_switch.hw_switch.number, coil.hw_driver.number)
        self._add_pulse_rule_to_switch(enable_switch, coil)
        self._add_release_disable_rule_to_switch(enable_switch, coil)

        self._write_rules_to_switch(enable_switch, coil, False)

    def set_pulse_on_hit_and_enable_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit and enable and relase rule on driver."""
        self.debug_log("Setting Pulse on hit and enable and release HW Rule. Switch: %s, Driver: %s",
                       enable_switch.hw_switch.number, coil.hw_driver.number)
        self._add_pulse_and_hold_rule_to_switch(enable_switch, coil)
        self._add_release_disable_rule_to_switch(enable_switch, coil)

        self._write_rules_to_switch(enable_switch, coil, False)

    def set_pulse_on_hit_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                      eos_switch: SwitchSettings, coil: DriverSettings,
                                                      repulse_settings: Optional[RepulseSettings]):
        """Set pulse on hit and enable and release and disable rule on driver."""
        self.debug_log("Setting Pulse on hit and release and disable HW Rule. Enable Switch: %s,"
                       "Disable Switch: %s, Driver: %s", enable_switch.hw_switch.number,
                       eos_switch.hw_switch.number, coil.hw_driver.number)
        self._add_pulse_rule_to_switch(enable_switch, coil)
        self._add_release_disable_rule_to_switch(enable_switch, coil)
        self._add_disable_rule_to_switch(eos_switch, coil)

        self._write_rules_to_switch(enable_switch, coil, False)
        self._write_rules_to_switch(eos_switch, coil, False)

    def set_pulse_on_hit_and_enable_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                                 eos_switch: SwitchSettings, coil: DriverSettings,
                                                                 repulse_settings: Optional[RepulseSettings]):
        """Set pulse on hit and enable and release and disable rule on driver."""
        self.debug_log("Setting Pulse on hit and enable and release and disable HW Rule. Enable Switch: %s,"
                       "Disable Switch: %s, Driver: %s", enable_switch.hw_switch.number,
                       eos_switch.hw_switch.number, coil.hw_driver.number)
        self._add_pulse_and_hold_rule_to_switch(enable_switch, coil)
        self._add_release_disable_rule_to_switch(enable_switch, coil)
        self._add_hold_rule_to_switch(eos_switch, coil)

        self._write_rules_to_switch(enable_switch, coil, False)
        self._write_rules_to_switch(eos_switch, coil, False)

    def clear_hw_rule(self, switch, coil):
        """Clear a hardware rule.

        This is used if you want to remove the linkage between a switch and
        some driver activity. For example, if you wanted to disable your
        flippers (so that a player pushing the flipper buttons wouldn't cause
        the flippers to flip), you'd call this method with your flipper button
        as the *sw_num*.

        Args:
        ----
            switch: Switch object
            coil: Coil object
        """
        self.debug_log("Clearing HW rule for switch: %s coil: %s", switch.hw_switch.number, coil.hw_driver.number)
        coil_number = False
        for entry, element in switch.hw_switch.hw_rules.items():
            if not element:
                continue
            for rule_num, rule in enumerate(element):
                if rule[0] == switch.hw_switch.number and rule[1] == coil.hw_driver.number:
                    coil_number = rule[2]['driverNum']
                    del switch.hw_switch.hw_rules[entry][rule_num]

        if coil_number:
            self.run_proc_cmd_no_wait("driver_disable", coil_number)
            self._write_rules_to_switch(switch, coil, None)

    def _get_default_subtype(self):
        """Return default subtype for either P3-Roc or P-Roc."""
        raise NotImplementedError

    def parse_light_number_to_channels(self, number: str, subtype: str):
        """Parse light number to a list of channels."""
        if not subtype:
            subtype = self._get_default_subtype()
        if subtype == "matrix":
            return [
                {
                    "number": number
                }
            ]
        if subtype == "led":
            # split the number (which comes in as a string like w-x-y-z) into parts
            number_parts = str(number).split('-')

            if len(number_parts) != 4:
                raise AssertionError("Invalid address for LED {}".format(number))

            return [
                {
                    "number": number_parts[0] + "-" + number_parts[1]
                },
                {
                    "number": number_parts[0] + "-" + number_parts[2]
                },
                {
                    "number": number_parts[0] + "-" + number_parts[3]
                },
            ]

        raise AssertionError("Unknown subtype {}".format(subtype))

    def configure_light(self, number, subtype, config, platform_settings) -> LightPlatformInterface:
        """Configure a light channel."""
        if not subtype:
            subtype = self._get_default_subtype()
        if subtype == "matrix":
            if self.machine_type == self.pinproc.MachineTypePDB:
                proc_num = self.pdbconfig.get_proc_light_number(str(number))
                if proc_num == -1:
                    raise AssertionError("Matrixlight {}/{} cannot be controlled by the P-ROC. ".format(
                        config.name, str(number)))

            else:
                proc_num = self.pinproc.decode(self.machine_type, str(number))

            return PROCMatrixLight(proc_num, self.machine, self)
        if subtype == "led":
            board, index = number.split("-")
            polarity = platform_settings and platform_settings.get("polarity", False)
            return PDBLED(int(board), int(index), polarity, self.config.get("debug", False), self, self._light_system)

        raise AssertionError("unknown subtype {}".format(subtype))

    def _configure_switch(self, config: SwitchConfig, proc_num) -> PROCSwitch:
        """Configure a P3-ROC switch.

        Args:
        ----
            config: Dictionary of settings for the switch.
            proc_num: decoded switch number

        Returns a reference to the switch object that was just created.
        """
        if proc_num == -1:
            raise AssertionError("Switch {}/{} cannot be controlled by the "
                                 "P-ROC/P3-ROC.".format(config.name, proc_num))

        switch = PROCSwitch(config, proc_num, config.debounce == "quick", self)
        # The P3-ROC needs to be configured to notify the host computers of
        # switch events. (That notification can be for open or closed,
        # debounced or nondebounced.)
        self.debug_log("Configuring switch's host notification settings. P3-ROC"
                       "number: %s, debounce: %s", proc_num,
                       config.debounce)
        if config.debounce == "quick":
            self.run_proc_cmd_no_wait("switch_update_rule", proc_num, 'closed_nondebounced',
                                      {'notifyHost': True, 'reloadActive': False}, [], False)
            self.run_proc_cmd_no_wait("switch_update_rule", proc_num, 'open_nondebounced',
                                      {'notifyHost': True, 'reloadActive': False}, [], False)
        else:
            self.run_proc_cmd_no_wait("switch_update_rule", proc_num, 'closed_debounced',
                                      {'notifyHost': True, 'reloadActive': False}, [], False)
            self.run_proc_cmd_no_wait("switch_update_rule", proc_num, 'open_debounced',
                                      {'notifyHost': True, 'reloadActive': False}, [], False)
        return switch

    async def configure_servo(self, number: str) -> ServoPlatformInterface:
        """Configure a servo on a PD-LED board.

        Args:
        ----
            number: Number of the servo
        """
        try:
            board, number = number.split("-")
        except ValueError:
            self.raise_config_error("Servo number should be board-number but is {}".format(number), 1)
        if 0 > int(number) >= 12:
            self.raise_config_error("PD-LED only supports 12 servos {}".format(number), 5)

        return PdLedServo(board, number, self, self.config.get("debug", False))

    async def configure_stepper(self, number: str, config: dict) -> PdLedStepper:
        """Configure a stepper (axis) device in platform.

        Args:
        ----
            number: Number of the stepper.
            config: Config for this stepper.
        """
        try:
            board, number = number.split("-")
        except ValueError:
            self.raise_config_error("Stepper number should be board-number but is {}".format(number), 3)
        if 0 > int(number) >= 2:
            self.raise_config_error("PD-LED only supports two steppers {}".format(number), 4)

        pd_led = self.config['pd_led_boards'].get(board, {})
        stepper_speed = pd_led.get("stepper_speed", 13524)

        return PdLedStepper(board, number, self, self.config.get("debug", False), stepper_speed)


class PDBConfig:

    """Handles PDB Config of the P/P3-Roc.

    This class is only used when using the P3-ROC or when the P-ROC is configured to use PDB
    driver boards such as the PD-16 or PD-8x8. i.e. not when it's operating in
    WPC or Stern mode.
    """

    __slots__ = ["log", "platform", "polarity", "lamp_matrix_strobe_time", "watchdog_time", "use_watchdog", "indexes"]

    # pylint: disable-msg=too-many-locals
    def __init__(self, proc_platform, config, driver_count):
        """Set up PDB config.

        Will configure driver groups for matrixes, lamps and normal drivers.
        We always use polarity True in this method because it is only used for PDB machines.
        """
        self.log = logging.getLogger('PDBConfig')
        self.log.debug("Processing PDB Driver Board configuration")

        self.platform = proc_platform

        # Set config defaults
        self.lamp_matrix_strobe_time = config['p_roc']['lamp_matrix_strobe_time']
        self.watchdog_time = config['p_roc']['watchdog_time']
        self.use_watchdog = config['p_roc']['use_watchdog']

        # Initialize some lists for data collecting
        coil_bank_list, unconfigured_coil_bank_list = self._load_coil_bank_list_from_config(config)
        lamp_source_bank_list, lamp_list, lamp_list_for_index = self._load_lamp_lists_from_config(config)

        # Create a list of indexes.  The PDB banks will be mapped into this
        # list. The index of the bank is used to calculate the P-ROC/P3-ROC driver
        # number for each driver.
        num_proc_banks = driver_count // 8
        self.indexes = [{}] * num_proc_banks    # type: List[Union[int, dict]]

        self._initialize_drivers()

        # Set up dedicated driver groups (groups 0-3).
        for group_ctr in range(0, 4):
            # PDB Banks 0-3 are interpreted as dedicated bank here. Therefore, we do not use them.
            enable = group_ctr in coil_bank_list
            self.log.debug("Driver group %02d (dedicated): Enable=%s",
                           group_ctr, enable)
            self.platform.run_proc_cmd_no_wait("driver_update_group_config",
                                               group_ctr,
                                               0,
                                               group_ctr,
                                               0,
                                               0,
                                               False,
                                               True,
                                               enable,
                                               True)

        # next group is 4
        group_ctr = 4

        # Process lamps first. The P-ROC/P3-ROC can only control so many drivers
        # directly. Since software won't have the speed to control lamp
        # matrixes, map the lamps first. If there aren't enough driver
        # groups for coils, the overflow coils can be controlled by software
        # via VirtualDrivers (which should get set up automatically by this
        # code.)
        lamp_banks = set()
        for i, lamp_dict in enumerate(lamp_list):
            # If the bank is 16 or higher, the P-ROC/P3-ROC can't control it
            # directly. Software can't really control lamp matrixes either
            # (need microsecond resolution).  Instead of doing crazy logic here
            # for a case that probably won't happen, just ignore these banks.
            if group_ctr >= num_proc_banks or lamp_dict['sink_bank'] >= 16:
                raise AssertionError("Lamp matrix banks can't be mapped to index "
                                     "{} because that's outside of the banks the "
                                     "P-ROC/P3-ROC can control.".format(lamp_dict['sink_bank']))

            self.log.debug("Driver group %02d (lamp sink): slow_time=%d "
                           "enable_index=%d row_activate_index=%d "
                           "row_enable_index=%d matrix=%s", group_ctr,
                           self.lamp_matrix_strobe_time,
                           lamp_dict['sink_bank'],
                           lamp_dict['source_output'],
                           lamp_dict['source_index'], True)
            self.indexes[group_ctr] = lamp_list_for_index[i]
            lamp_banks.add(lamp_dict['sink_bank'])
            self.platform.run_proc_cmd_no_wait("driver_update_group_config",
                                               group_ctr,
                                               self.lamp_matrix_strobe_time,
                                               lamp_dict['sink_bank'],
                                               lamp_dict['source_output'],
                                               lamp_dict['source_index'],
                                               True,
                                               True,
                                               True,
                                               True)
            group_ctr += 1

        unconfigured_coil_bank_list -= lamp_banks
        unconfigured_coil_bank_list -= set(lamp_source_bank_list)

        for coil_bank in coil_bank_list:
            # If the bank is 16 or higher, the P-ROC/P3-ROC can't control it directly.
            # Software will have do the driver logic and write any changes to
            # the PDB bus. Therefore, map these banks to indexes above the
            # driver count, which will force the drivers to be created
            # as VirtualDrivers. Appending the bank avoids conflicts when
            # group_ctr gets too high.
            if group_ctr >= num_proc_banks or coil_bank >= 32:
                self.log.warning("Driver group %d mapped to driver index"
                                 "outside of P-ROC/P3-ROC control.  These Drivers "
                                 "will become VirtualDrivers.  Note, the "
                                 "index will not match the board/bank "
                                 "number; so software will need to request "
                                 "those values before updating the "
                                 "drivers.", coil_bank)
                self.indexes.append(coil_bank)
            else:
                self.log.debug("Driver group %02d: slow_time=%d Enable "
                               "Index=%d", group_ctr, 0, coil_bank)
                self.indexes[group_ctr] = coil_bank
                self.platform.run_proc_cmd_no_wait("driver_update_group_config",
                                                   group_ctr,
                                                   0,
                                                   coil_bank,
                                                   0,
                                                   0,
                                                   False,
                                                   True,
                                                   True,
                                                   True)
                group_ctr += 1

        # configure all unconfigured coil banks but do not enable them
        for coil_bank in unconfigured_coil_bank_list:
            if group_ctr >= num_proc_banks or coil_bank >= 32:
                self.log.warning("Cannot configure %s. The polarity on those banks might be incorrect.", coil_bank)
                continue

            self.log.debug("Driver group %02d: slow_time=%d Enabled "
                           "Index=%d", group_ctr, 0, coil_bank)
            self.platform.run_proc_cmd_no_wait("driver_update_group_config",
                                               group_ctr,
                                               0,
                                               coil_bank,
                                               0,
                                               0,
                                               False,
                                               True,
                                               True,
                                               True)
            group_ctr += 1

        for i in range(group_ctr, 26):
            self.log.debug("Driver group %02d: disabled", i)
            self.platform.run_proc_cmd_no_wait("driver_update_group_config",
                                               i,
                                               self.lamp_matrix_strobe_time,
                                               0,
                                               0,
                                               0,
                                               False,
                                               True,
                                               False,
                                               True)

        # Make sure there are two indexes.  If not, fill them in.
        while len(lamp_source_bank_list) < 2:
            lamp_source_bank_list.append(0)

        # Now set up globals.  First disable them to allow the P-ROC/P3-ROC to set up
        # the polarities on the Drivers.  Then enable them.
        self._configure_lamp_banks(lamp_source_bank_list, False)
        self._configure_lamp_banks(lamp_source_bank_list, True)

    def is_pdb_address(self, addr):
        """Return True if the given address is a valid PDB address."""
        try:
            self.decode_pdb_address(addr=addr)
            return True
        except ValueError:
            return False

    @staticmethod
    def decode_pdb_address(addr):
        """Decode Ax-By-z or x/y/z into PDB address, bank number, and output number.

        Raises a ValueError exception if it is not a PDB address, otherwise returns
        a tuple of (addr, bank, number).

        """
        if '-' in addr:  # Ax-By-z form
            params = addr.rsplit('-')
            if len(params) != 3:
                raise ValueError('pdb address must have 3 components')
            board = int(params[0][1:])
            bank = int(params[1][1:])
            output = int(params[2][0:])
            return board, bank, output

        if '/' in addr:  # x/y/z form
            params = addr.rsplit('/')
            if len(params) != 3:
                raise ValueError('pdb address must have 3 components')
            board = int(params[0])
            bank = int(params[1])
            output = int(params[2])
            return board, bank, output

        raise ValueError('PDB address delimiter (- or /) not found.')

    def _load_lamp_lists_from_config(self, config):
        lamp_source_bank_list = []
        lamp_list = []
        lamp_list_for_index = []

        # Make a list of unique lamp source banks.  The P-ROC/P3-ROC only supports 2.
        # If this is exceeded we will error out later.
        if 'lights' in config:
            for name in config['lights']:
                item_dict = config['lights'][name]
                if "subtype" not in item_dict or item_dict["subtype"] != "matrix":
                    continue
                lamp = PDBLight(self, str(item_dict['number']))

                # Catalog PDB banks
                # Dedicated lamps don't use PDB banks. They use P-ROC direct
                # driver pins (not available on P3-ROC).
                if lamp.lamp_type == 'dedicated':
                    pass

                elif lamp.lamp_type == 'pdb':
                    if lamp.source_bank() not in lamp_source_bank_list:
                        lamp_source_bank_list.append(lamp.source_bank())

                    # Create dicts of unique sink banks.  The source index is
                    # needed when setting up the driver groups.
                    lamp_dict = {'source_index':
                                 lamp_source_bank_list.index(lamp.source_bank()),
                                 'sink_bank': lamp.sink_bank(),
                                 'source_output': lamp.source_output()}

                    # lamp_dict_for_index.  This will be used later when the
                    # p-roc numbers are requested.  The requester won't know
                    # the source_index, but it will know the source board.
                    # This is why two separate lists are needed.
                    lamp_dict_for_index = {'source_board': lamp.source_board(),
                                           'sink_bank': lamp.sink_bank(),
                                           'source_output':
                                               lamp.source_output()}

                    if lamp_dict not in lamp_list:
                        lamp_list.append(lamp_dict)
                        lamp_list_for_index.append(lamp_dict_for_index)

        return lamp_source_bank_list, lamp_list, lamp_list_for_index

    def _load_coil_bank_list_from_config(self, config) -> Tuple[Set[int], Set[int]]:
        coil_bank_list = set()
        secondary_coil_bank_list = set()

        # Make a list of unique coil banks
        if 'coils' in config:
            for name in config['coils']:
                item_dict = config['coils'][name]
                coil = PDBCoil(self, str(item_dict['number']))
                bank = coil.bank()
                coil_bank_list.add(bank)

                secondary_bank = coil.bank_secondary()
                if secondary_bank is not None:
                    secondary_coil_bank_list.add(secondary_bank)

        secondary_coil_bank_list -= coil_bank_list

        return coil_bank_list, secondary_coil_bank_list

    def _initialize_drivers(self):
        """Loop through all of the drivers, initializing them with the polarity."""
        for i in range(0, 255):
            state = {'driverNum': i,
                     'outputDriveTime': 0,
                     'polarity': True,
                     'state': False,
                     'waitForFirstTimeSlot': False,
                     'timeslots': 0,
                     'patterOnTime': 0,
                     'patterOffTime': 0,
                     'patterEnable': False,
                     'futureEnable': False}

            self.platform.run_proc_cmd_no_wait("driver_update_state", state)

    def _configure_lamp_banks(self, lamp_source_bank_list, enable):
        self.platform.run_proc_cmd_no_wait("driver_update_global_config",
                                           enable,
                                           True,  # Polarity
                                           False,  # N/A
                                           False,  # N/A
                                           1,  # N/A
                                           lamp_source_bank_list[0],
                                           lamp_source_bank_list[1],
                                           False,  # Active low rows? No
                                           False,  # N/A
                                           False,  # Stern? No
                                           False,  # Reset watchdog trigger
                                           self.use_watchdog,  # Enable watchdog
                                           self.watchdog_time)

        if enable:
            self.log.debug("Configuring PDB Driver Globals:  polarity = %s  "
                           "matrix column index 0 = %d  matrix column index "
                           "1 = %d", True, lamp_source_bank_list[0],
                           lamp_source_bank_list[1])

    def get_proc_coil_number(self, number_str):
        """Get the actual number of a coil from the bank index config.

        Args:
        ----
            number_str (str): PDB string
        """
        coil = PDBCoil(self, number_str)
        bank = coil.bank()
        if bank == -1:
            return -1
        index = self.indexes.index(coil.bank())
        num = index * 8 + coil.output()
        return num

    def get_proc_light_number(self, number_str):
        """Get the actual number of a light from the lamp config.

        Args:
        ----
            number_str (str): PDB string
        """
        lamp = PDBLight(self, number_str)
        if lamp.lamp_type == 'unknown':
            return -1
        if lamp.lamp_type == 'dedicated':
            return lamp.dedicated_output()

        lamp_dict_for_index = {'source_board': lamp.source_board(),
                               'sink_bank': lamp.sink_bank(),
                               'source_output': lamp.source_output()}
        if lamp_dict_for_index not in self.indexes:
            raise AssertionError("Light not in lamp dict")
        index = self.indexes.index(lamp_dict_for_index)
        num = index * 8 + lamp.sink_output()
        return num

    def get_proc_switch_number(self, number_str):
        """Get the actual number of a switch based on the string only.

        Args:
        ----
            number_str (str): PDB string
        """
        switch = PDBSwitch(self, number_str)
        num = switch.proc_num()
        return num



"""OPP WS2812 wing."""
import logging

from mpf.core.platform_batch_light_system import PlatformBatchLight

from mpf.platforms.opp.opp_rs232_intf import OppRs232Intf


class OPPNeopixelCard:

    """OPP Neopixel/WS281x card."""

    __slots__ = ["log", "chain_serial", "platform", "addr", "card_num"]

    def __init__(self, chain_serial, addr, platform):
        """Initialise OPP Neopixel/WS2812 card."""
        self.log = logging.getLogger('OPPNeopixel {} on {}'.format(addr, chain_serial))
        self.chain_serial = chain_serial
        self.addr = addr
        self.platform = platform
        self.card_num = str(addr - ord(OppRs232Intf.CARD_ID_GEN2_CARD))

        self.log.debug("Creating OPP Neopixel card at hardware address: 0x%02x", addr)

    @staticmethod
    def is_valid_light_number(number):
        """Check if neopixel number is possible in hardware."""
        return 0 <= int(number) < 0x1000


class OPPModernMatrixLightsCard:

    """OPP Matrix Lights Card with firmware >= 2.1.0."""

    __slots__ = ["log", "chain_serial", "platform", "addr", "card_num"]

    def __init__(self, chain_serial, addr, platform):
        """Initialise OPP Incand card."""
        self.log = logging.getLogger('OPPMatrixLights {} on {}'.format(addr, chain_serial))
        self.chain_serial = chain_serial
        self.addr = addr
        self.platform = platform
        self.card_num = str(addr - ord(OppRs232Intf.CARD_ID_GEN2_CARD))

        self.log.debug("Creating OPP Matrix Lights card at hardware address: 0x%02x", addr)

    @staticmethod
    def is_valid_light_number(number):
        """Check if matrix light exists in hardware."""
        return 0 <= int(number) < 64


class OPPModernLightChannel(PlatformBatchLight):

    """A channel of a WS2812 LED."""

    __slots__ = ["chain_serial", "addr", "pixel_num"]

    # pylint: disable-msg=too-many-arguments
    def __init__(self, chain_serial, addr, pixel_num, light_system):
        """Initialise led channel."""
        super().__init__("{}-{}-{}".format(chain_serial, addr, pixel_num), light_system)
        self.pixel_num = pixel_num
        self.addr = addr
        self.chain_serial = chain_serial

    def get_max_fade_ms(self):
        """Return largest number which fits two bytes."""
        return 65535

    def get_type_string(self):
        """Return string for the type."""
        if self.pixel_num < 0x1000:
            return "LED"

        if self.pixel_num < 0x2000:
            return "Incandescent Lamp"

        return "Matrix Lamp"

    def get_board_name(self):
        """Return OPP chain and addr."""
        return "OPP Light {} on Chain {} Board {} Type {}".format(
            self.pixel_num, self.chain_serial, self.addr, self.get_type_string())

    def is_successor_of(self, other):
        """Return true if the other light has the previous pixel_num and is on the same chain and addr."""
        return (self.chain_serial == other.chain_serial and self.addr == other.addr and
                self.pixel_num == other.pixel_num + 1)

    def get_successor_number(self):
        """Return nex pixel_num on the same chain and addr."""
        return "{}-{}-{}".format(self.chain_serial, self.addr, self.pixel_num + 1)

    def __lt__(self, other):
        """Order lights by their position on the hardware."""
        return (self.chain_serial, self.addr, self.pixel_num) < (other.chain_serial, other.addr, other.pixel_num)

    def __repr__(self):
        """Return str representation."""
        return "<OPPLightChannel type={} chain={} addr={} pixel={}>".format(
            self.get_type_string(), self.chain_serial, self.addr, self.pixel_num)



"""Contains the ConfigProcessor."""

import errno
import hashlib
import logging
import os
import pickle   # nosec
import tempfile

from typing import List, Tuple, Any, Optional
from difflib import SequenceMatcher
from pkg_resources import iter_entry_points

from mpf.file_interfaces.yaml_interface import YamlInterface
import mpf
from mpf.core.file_manager import FileManager
from mpf.core.utility_functions import Util
from mpf._version import __show_version__, __config_version__
from mpf.exceptions.config_file_error import ConfigFileError


class ConfigProcessor:

    """Config processor which loads the config."""

    __slots__ = ["log", "_load_cache", "_store_cache"]

    def __init__(self, load_cache, store_cache):
        """Initialise config processor."""
        self.log = logging.getLogger("ConfigProcessor")
        self._load_cache = load_cache
        self._store_cache = store_cache

    @staticmethod
    def get_cache_dir():
        """Return cache dir."""
        return tempfile.gettempdir()

    def get_cache_filename(self, filenames: List[str]) -> str:   # pragma: no cover
        """Return cache file name."""
        cache_dir = self.get_cache_dir()
        filestring = ""
        for configfile in filenames:
            filestring += str(os.path.abspath(configfile))
        path_hash = hashlib.md5(bytes(filestring, 'UTF-8')).hexdigest()     # nosec
        return os.path.join(cache_dir, path_hash + ".mpf_cache")

    def _load_config_from_cache(self, cache_file) -> Tuple[Any, Optional[List[str]]]:     # nosec
        """Return config from cache."""
        self.log.info("Loading config from cache: %s", cache_file)
        with open(cache_file, 'rb') as f:
            try:
                data = pickle.load(f)   # type: Tuple[Any, List[str]]
            # unfortunately pickle can raise all kinds of exceptions and we dont want to crash on corrupted cache
            # pylint: disable-msg=broad-except
            except Exception:   # pragma: no cover
                self.log.warning("Could not load cache file: %s", cache_file)
                return None, None

        if not isinstance(data, tuple) or len(data) != 2:
            return None, None
        return data

    def _get_mtime_or_negative(self, filename) -> float:
        """Return mtime or -1 if file could not be found."""
        try:
            return os.path.getmtime(filename)
        except OSError as exception:
            if exception.errno != errno.ENOENT:
                raise  # some unknown error?

            self.log.debug('Cache file not found: %s', filename)
            return -1

    # pylint: disable-msg=too-many-arguments
    def load_config_files_with_cache(self, filenames: List[str], config_type: str,
                                     ignore_unknown_sections=False, config_spec=None) -> dict:   # pragma: no cover
        """Load multiple configs with a combined cache."""
        config = dict()     # type: Any
        # Step 1: Check timestamps of the filelist vs cache
        cache_file = self.get_cache_filename(filenames)
        load_from_cache = self._load_cache
        if load_from_cache:
            cache_time = self._get_mtime_or_negative(cache_file)
            if cache_time < 0:
                load_from_cache = False
            else:
                for configfile in filenames:
                    if not os.path.isfile(configfile) or os.path.getmtime(configfile) > cache_time:
                        load_from_cache = False
                        self.log.warning('Config file in cache changed: %s', configfile)
                        break

        # Step 2: Get cache content
        if load_from_cache:
            config, loaded_files = self._load_config_from_cache(cache_file)
            if not config:
                load_from_cache = False
        else:
            loaded_files = []

        # Step 3: Check timestamps of included files vs cache
        if loaded_files:
            for configfile in loaded_files:
                if not os.path.isfile(configfile) or os.path.getmtime(configfile) > cache_time:
                    load_from_cache = False
                    self.log.warning('Config file in cache changed: %s', configfile)
                    break

        # Step 4: Return cache
        if load_from_cache:
            return config

        config = dict()
        loaded_files = []
        for configfile in filenames:
            self.log.info('Loading config from file %s.', configfile)
            file_config, file_subfiles = self._load_config_file_and_return_loaded_files(configfile, config_type,
                                                                                        ignore_unknown_sections,
                                                                                        config_spec=config_spec)
            loaded_files.extend(file_subfiles)
            config = Util.dict_merge(config, file_config)

        # Step 5: Store to cache
        if self._store_cache:
            with open(cache_file, 'wb') as f:
                pickle.dump((config, loaded_files), f, protocol=4)
                self.log.info('Config file cache created: %s', cache_file)

        return config

    @staticmethod
    def _find_similar_key(key, config_spec, config_type):
        """Find the most similar key in config_spec."""
        best_similarity = 0
        best_key = ""
        for existing_key, config in config_spec.items():
            if '__valid_in__' not in config or config_type not in config['__valid_in__']:
                continue
            similarity = SequenceMatcher(None, existing_key, key).ratio()
            if similarity > best_similarity:
                best_key = existing_key
                best_similarity = similarity

        return best_key

    def _check_sections(self, config_spec, config, config_type, filename, ignore_unknown_sections):
        if not isinstance(config, dict):
            raise ConfigFileError("Config should be a dict: {}".format(config), 1, self.log.name, filename)
        for k in config.keys():
            if k in config_spec:
                if config_type not in config_spec[k]['__valid_in__']:
                    raise ConfigFileError('Found a "{}:" section in config file {}, '
                                          'but that section is not valid in {} config '
                                          'files (only in {}).'.format(k, filename, config_type,
                                                                       config_spec[k]['__valid_in__']),
                                          2, self.log.name, filename)
            elif not ignore_unknown_sections:
                suggestion = self._find_similar_key(k, config_spec, config_type)

                raise ConfigFileError('Found a "{}:" section in config file {}, '
                                      'but that section name is unknown. Did you mean "{}:" '
                                      'instead?.'.format(k, filename, suggestion), 3, self.log.name,
                                      filename)

    def _load_config_file_and_return_loaded_files(
            self, filename, config_type: str,
            ignore_unknown_sections=False, config_spec=None) -> Tuple[dict, List[str]]:   # pragma: no cover
        """Load a config file and return loaded files."""
        # config_type is str 'machine' or 'mode', which specifies whether this
        # file being loaded is a machine config or a mode config file
        expected_version_str = ConfigProcessor.get_expected_version(config_type)

        config = FileManager.load(filename, expected_version_str, True)
        subfiles = []

        if not config:
            return dict(), []

        self.log.info('Loading config: %s', filename)

        if config_type in ("machine", "mode"):
            self._check_sections(config_spec, config, config_type, filename, ignore_unknown_sections)

        try:
            if 'config' in config:
                path = os.path.split(filename)[0]

                for file in Util.string_to_event_list(config['config']):
                    full_file = os.path.join(path, file)
                    subfiles.append(full_file)
                    subconfig, subsubfiles = self._load_config_file_and_return_loaded_files(full_file, config_type,
                                                                                            config_spec=config_spec)
                    subfiles.extend(subsubfiles)
                    config = Util.dict_merge(config, subconfig)
            return config, subfiles
        except TypeError:
            return dict(), []

    def load_config_spec(self):
        """Load config spec."""
        cache_file = os.path.join(self.get_cache_dir(), "config_spec.mpf_cache")
        config_spec_file = os.path.abspath(os.path.join(mpf.core.__path__[0], os.pardir, "config_spec.yaml"))
        stats_config_spec_file = os.stat(config_spec_file)
        if self._load_cache and os.path.isfile(cache_file) and \
                os.path.getmtime(cache_file) == stats_config_spec_file.st_mtime:
            try:
                with open(cache_file, 'rb') as f:
                    return pickle.load(f)   # nosec
            except Exception:   # noqa
                pass

        config = FileManager.load(config_spec_file, False, True)
        config = self._process_config_spec(config, "root")

        config = self.load_external_platform_config_specs(config)

        if self._store_cache:
            with open(cache_file, 'wb') as f:
                pickle.dump(config, f, protocol=4)
                os.utime(cache_file, ns=(stats_config_spec_file.st_atime_ns, stats_config_spec_file.st_mtime_ns))
                self.log.info('Config spec file cache created: %s', cache_file)

        return config

    def load_external_platform_config_specs(self, config):
        """Load config spec for external platforms."""
        for platform_entry in iter_entry_points(group='mpf.platforms'):
            config_spec = platform_entry.load().get_config_spec()

            if config_spec:
                # add specific config spec if platform has any
                config[config_spec[1]] = self._process_config_spec(YamlInterface.process(config_spec[0]),
                                                                   config_spec[1])
        return config

    def load_device_config_specs(self, config_spec, machine_config):
        """Load device config specs."""
        for device_type in machine_config['mpf']['device_modules']:
            device_cls = Util.string_to_class(device_type)
            if device_cls.get_config_spec():
                # add specific config spec if device has any
                config_spec[device_cls.config_section] = self._process_config_spec(
                    YamlInterface.process(device_cls.get_config_spec()),
                    device_cls.config_section)

        return config_spec

    def _process_config_spec(self, spec, path):
        if not isinstance(spec, dict):
            raise AssertionError("Expected a dict at: {} {}".format(path, spec))

        for element, value in spec.items():
            if element.startswith("__"):
                spec[element] = value
            elif isinstance(value, str):
                if value == "ignore":
                    spec[element] = value
                else:
                    spec[element] = value.split('|')
                    if len(spec[element]) != 3:
                        raise AssertionError("Format incorrect: {}".format(value))
            else:
                spec[element] = self._process_config_spec(value, path + ":" + element)

        return spec

    @staticmethod
    def get_expected_version(config_type: str) -> str:
        """Return the expected config or show version tag, e.g. #config_version=5."""
        if config_type in ("machine", "mode"):
            return "#config_version={}".format(__config_version__)
        if config_type == "show":
            return "#show_version={}".format(__show_version__)

        raise AssertionError("Invalid config_type {}".format(config_type))



"""Contains the parent class for all platforms."""
import abc
import asyncio
from collections import namedtuple
from enum import Enum

from typing import Optional, Dict, List

from mpf.core.logging import LogMixin
from mpf.core.utility_functions import Util

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.devices.switch import Switch   # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.stepper import Stepper     # pylint: disable-msg=cyclic-import,unused-import
    from mpf.platforms.interfaces.driver_platform_interface import DriverPlatformInterface  # pylint: disable-msg=cyclic-import,unused-import; # noqa
    from mpf.platforms.interfaces.switch_platform_interface import SwitchPlatformInterface  # pylint: disable-msg=cyclic-import,unused-import; # noqa
    from mpf.platforms.interfaces.light_platform_interface import LightPlatformInterface    # pylint: disable-msg=cyclic-import,unused-import; # noqa
    from mpf.platforms.interfaces.servo_platform_interface import ServoPlatformInterface    # pylint: disable-msg=cyclic-import,unused-import; # noqa
    from mpf.platforms.interfaces.segment_display_platform_interface import SegmentDisplayPlatformInterface     # pylint: disable-msg=cyclic-import,unused-import; # noqa
    from mpf.platforms.interfaces.hardware_sound_platform_interface import HardwareSoundPlatformInterface   # pylint: disable-msg=cyclic-import,unused-import; # noqa
    from mpf.platforms.interfaces.stepper_platform_interface import StepperPlatformInterface    # pylint: disable-msg=cyclic-import,unused-import; # noqa
    from mpf.platforms.interfaces.accelerometer_platform_interface import AccelerometerPlatformInterface    # pylint: disable-msg=cyclic-import,unused-import; # noqa
    from mpf.platforms.interfaces.i2c_platform_interface import I2cPlatformInterface    # pylint: disable-msg=cyclic-import,unused-import; # noqa
    from mpf.platforms.interfaces.dmd_platform import DmdPlatformInterface  # pylint: disable-msg=cyclic-import,unused-import; # noqa
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import; # noqa


class BasePlatform(LogMixin, metaclass=abc.ABCMeta):

    """Base class for all hardware platforms in MPF."""

    __slots__ = ["machine", "features", "debug"]

    def __init__(self, machine):
        """Create features and set default variables.

        Args:
        ----
            machine(mpf.core.machine.MachineController): The machine.
        """
        self.machine = machine  # type: MachineController
        self.features = {}
        super().__init__()
        self.debug = False

        # Set default platform features. Each platform interface can change
        # these to notify the framework of the specific features it supports.
        self.features['has_dmds'] = False
        self.features['has_rgb_dmds'] = False
        self.features['has_accelerometers'] = False
        self.features['has_i2c'] = False
        self.features['has_servos'] = False
        self.features['has_lights'] = False
        self.features['has_switches'] = False
        self.features['has_drivers'] = False
        self.features['tickless'] = False
        self.features['has_segment_displays'] = False
        self.features['has_hardware_sound_systems'] = False
        self.features['has_steppers'] = False
        self.features['allow_empty_numbers'] = False
        self.features['hardware_eos_repulse'] = False

    def assert_has_feature(self, feature_name):
        """Assert that this platform has a certain feature or raise an exception otherwise."""
        if not self.features.get("has_{}".format(feature_name), False):
            self.raise_config_error("Platform {} does not support to configure {feature_name}. "
                                    "Please make sure the platform "
                                    "you configured for {feature_name} actually supports that type "
                                    "of devices.".format(self.__class__, feature_name=feature_name), 99)

    def _configure_device_logging_and_debug(self, logger_name, config):
        """Configure logging for platform."""
        if config['debug']:
            self.debug = True
            config['console_log'] = 'full'
            config['file_log'] = 'full'

        self.configure_logging(logger_name,
                               config['console_log'],
                               config['file_log'])

    @classmethod
    def get_config_spec(cls):
        """Return config spec for this platform."""
        return False

    # pylint: disable-msg=no-self-use
    def get_info_string(self) -> str:
        """Return information string about this platform."""
        return "Not implemented"

    # pylint: disable-msg=no-self-use
    def update_firmware(self) -> str:
        """Perform a firmware update."""

    async def initialize(self):
        """Initialise the platform.

        This is called after all platforms have been created and core modules have been loaded.
        """

    async def start(self):
        """Start receiving switch changes from this platform."""

    def tick(self):
        """Run task.

        Called periodically.

        Subclass this method in a platform module to perform periodic updates
        to the platform hardware, e.g. reading switches, sending driver or
        light updates, etc.

        """

    def stop(self):
        """Stop the platform.

        Subclass this method in the platform module if you need to perform
        any actions to gracefully stop the platform interface.

        This could do things like reseting it, stopping events, etc.

        This method will be called when MPF stops, including when an MPF thread
        crashes.

        """


class DmdPlatform(BasePlatform, metaclass=abc.ABCMeta):

    """Baseclass for DMDs in MPF."""

    __slots__ = []  # type: List[str]

    def __init__(self, machine):
        """Add dmd feature."""
        super().__init__(machine)
        self.features['has_dmds'] = True

    @abc.abstractmethod
    def configure_dmd(self) -> "DmdPlatformInterface":
        """Subclass this method in a platform module to configure the DMD.

        This method should return a reference to the DMD's platform interface
        method will will receive the frame data.

        """
        raise NotImplementedError


class HardwareSoundPlatform(BasePlatform, metaclass=abc.ABCMeta):

    """Baseclass for hardware sounds in MPF."""

    __slots__ = []  # type: List[str]

    def __init__(self, machine):
        """Add hardware sound feature."""
        super().__init__(machine)
        self.features['has_hardware_sound_systems'] = True

    @abc.abstractmethod
    def configure_hardware_sound_system(self) -> "HardwareSoundPlatformInterface":
        """Return a reference to the hardware sound interface."""
        raise NotImplementedError


class RgbDmdPlatform(BasePlatform, metaclass=abc.ABCMeta):

    """Baseclass for RGB DMDs in MPF."""

    __slots__ = []  # type: List[str]

    def __init__(self, machine):
        """Add rgb dmd feature."""
        super().__init__(machine)
        self.features['has_rgb_dmds'] = True

    @abc.abstractmethod
    def configure_rgb_dmd(self, name: str) -> "DmdPlatformInterface":
        """Subclass this method in a platform module to configure the DMD.

        This method should return a reference to the DMD's platform interface
        method will will receive the frame data.

        """
        raise NotImplementedError


class SegmentDisplayPlatform(BasePlatform, metaclass=abc.ABCMeta):

    """Baseclass for 7-segment/6-digits display in MPF."""

    __slots__ = []  # type: List[str]

    def __init__(self, machine):
        """Add segment display feature."""
        super().__init__(machine)
        self.features['has_segment_displays'] = True

    @abc.abstractmethod
    async def configure_segment_display(self, number: str, platform_settings) -> "SegmentDisplayPlatformInterface":
        """Subclass this method in a platform module to configure a segment display.

        This method should return a reference to the segment display platform interface
        method will will receive the text to show.
        """
        raise NotImplementedError


# pylint: disable-msg=abstract-method
class SegmentDisplaySoftwareFlashPlatform(SegmentDisplayPlatform, metaclass=abc.ABCMeta):

    """SegmentDisplayPlatform with software flash support."""

    def __init__(self, machine):
        """Initialise software flash support."""
        super().__init__(machine)
        self._displays = set()
        self._display_flash_task = None

    async def initialize(self):
        """Start flash task."""
        await super().initialize()
        self._display_flash_task = self.machine.clock.loop.create_task(self._display_flash())
        self._display_flash_task.add_done_callback(Util.raise_exceptions)

    async def _display_flash(self):
        wait_time = 1 / (self.config['display_flash_frequency'] * 2)
        while True:
            # set on
            await asyncio.sleep(wait_time)
            for display in self._displays:
                display.set_software_flash(True)
            # set off
            await asyncio.sleep(wait_time)
            for display in self._displays:
                display.set_software_flash(False)

    def stop(self):
        """Cancel flash task."""
        super().stop()
        if self._display_flash_task:
            self._display_flash_task.cancel()

    def _handle_software_flash(self, display):
        """Register display for flash task."""
        self._displays.add(display)


class AccelerometerPlatform(BasePlatform, metaclass=abc.ABCMeta):

    """Baseclass for Accelerometer platforms."""

    __slots__ = []  # type: List[str]

    def __init__(self, machine):
        """Add accelerometer feature."""
        super().__init__(machine)
        self.features['has_accelerometers'] = True

    @abc.abstractmethod
    def configure_accelerometer(self, number: str, config: dict, callback) -> "AccelerometerPlatformInterface":
        """Configure accelerometer.

        Args:
        ----
            number: Number of this accelerometer
            config (dict): Configuration of this accelerometer
            callback (mpf.devices.accelerometer.Accelerometer): Callback device to send data to
        """
        raise NotImplementedError


class I2cPlatform(BasePlatform, metaclass=abc.ABCMeta):

    """Baseclass for I2C Platforms."""

    __slots__ = []  # type: List[str]

    def __init__(self, machine):
        """Initialise I2C platform and set feature."""
        super().__init__(machine)
        self.features['has_i2c'] = True

    async def configure_i2c(self, number: str) -> "I2cPlatformInterface":
        """Configure i2c device."""
        raise NotImplementedError


class ServoPlatform(BasePlatform, metaclass=abc.ABCMeta):

    """Baseclass for servo platforms in MPF."""

    __slots__ = []  # type: List[str]

    def __init__(self, machine):
        """Add servo feature."""
        super().__init__(machine)
        self.features['has_servos'] = True

    @abc.abstractmethod
    async def configure_servo(self, number: str) -> "ServoPlatformInterface":
        """Configure a servo device in platform.

        Args:
        ----
            number: Number of the servo
        """
        raise NotImplementedError


class StepperPlatform(BasePlatform, metaclass=abc.ABCMeta):

    """Baseclass for smart servo (axis) platforms in MPF."""

    __slots__ = []  # type: List[str]

    def __init__(self, machine):
        """Add smart servo feature."""
        super().__init__(machine)
        self.features['has_steppers'] = True

    @classmethod
    def get_stepper_config_section(cls) -> Optional[str]:
        """Return config section for additional stepper config items."""
        return None

    def validate_stepper_section(self, stepper: "Stepper", config: dict) -> dict:
        """Validate a stepper config for platform.

        Args:
        ----
            stepper: Stepper to validate.
            config: Config to validate.

        Returns: Validated config.
        """
        if self.get_stepper_config_section():
            spec = self.get_stepper_config_section()    # pylint: disable-msg=assignment-from-none
            config = stepper.machine.config_validator.validate_config(spec, config, stepper.name)
        elif config:
            raise AssertionError("No platform_config supported but not empty {} for stepper {}".
                                 format(config, stepper.name))

        return config

    @abc.abstractmethod
    async def configure_stepper(self, number: str, config: dict) -> "StepperPlatformInterface":
        """Configure a smart stepper (axis) device in platform.

        Args:
        ----
            number: Number of the smart servo
            config: Config for this stepper.
        """
        raise NotImplementedError


class LightConfigColors(Enum):

    """Light color for LightConfig."""

    RED = 1
    GREEN = 2
    BLUE = 3
    WHITE = 4
    NONE = 5


LightConfig = namedtuple("LightConfig", ["name", "color"])


class LightsPlatform(BasePlatform, metaclass=abc.ABCMeta):

    """Baseclass for platforms with any kind of lights in MPF.

    This includes LEDs, GIs, Matrix Lights and any other lights.
    """

    __slots__ = []  # type: List[str]

    def __init__(self, machine):
        """Add led feature."""
        super().__init__(machine)
        self.features['has_lights'] = True

    @abc.abstractmethod
    def parse_light_number_to_channels(self, number: str, subtype: str):
        """Parse light number to a list of channels."""
        raise NotImplementedError

    def light_sync(self):
        """Update lights synchronously.

        Called after channels of a light were updated. Can be used if multiple channels need to be flushed at once.
        """

    @abc.abstractmethod
    def configure_light(self, number: str, subtype: str, config: LightConfig,
                        platform_settings: dict) -> "LightPlatformInterface":
        """Subclass this method in a platform module to configure a light.

        This method should return a reference to the light
        object which will be called to access the hardware.
        """
        raise NotImplementedError


SwitchConfig = namedtuple("SwitchConfig", ["name", "invert", "debounce"])


class SwitchPlatform(BasePlatform, metaclass=abc.ABCMeta):

    """Baseclass for platforms with switches in MPF."""

    __slots__ = []  # type: List[str]

    def __init__(self, machine):
        """Add switch feature."""
        super().__init__(machine)
        self.features['has_switches'] = True

    @abc.abstractmethod
    def configure_switch(self, number: str, config: SwitchConfig, platform_config: dict) -> "SwitchPlatformInterface":
        """Subclass this method in a platform module to configure a switch.

        This method should return a reference to the switch's platform interface
        object which will be called to access the hardware.

        Args:
        ----
            number: Switch number.
            config : Config of switch.
            platform_config: Platform specific settings.

        """
        raise NotImplementedError

    @classmethod
    def get_switch_config_section(cls) -> Optional[str]:
        """Return config section for additional switch config items."""
        return None

    def validate_switch_section(self, switch: "Switch", config: dict) -> dict:
        """Validate a switch config for platform.

        Args:
        ----
            switch: Switch to validate.
            config: Config to validate.

        Returns: Validated config.
        """
        if self.get_switch_config_section():
            spec = self.get_switch_config_section()     # pylint: disable-msg=assignment-from-none
            config = switch.machine.config_validator.validate_config(spec, config, switch.name)
        elif config:
            raise AssertionError("No platform_config supported but not empty {} for switch {}".
                                 format(config, switch.name))

        return config

    @abc.abstractmethod
    async def get_hw_switch_states(self) -> Dict[str, bool]:
        """Get all hardware switch states.

        Subclass this method in a platform module to return the hardware
        states of all the switches on that platform.
        of a switch.

        This method should return a dict with the switch numbers as keys and the
        hardware state of the switches as values. (0 = inactive, 1 = active)
        This method should not compensate for NO or NC status, rather, it
        should return the raw hardware states of the switches.

        """
        raise NotImplementedError


SwitchSettings = namedtuple("SwitchSettings", ["hw_switch", "invert", "debounce"])
DriverSettings = namedtuple("DriverSettings", ["hw_driver", "pulse_settings", "hold_settings", "recycle"])
DriverConfig = namedtuple("DriverConfig", ["name", "default_pulse_ms", "default_pulse_power", "default_hold_power",
                                           "default_recycle", "max_pulse_ms", "max_pulse_power", "max_hold_power"])
RepulseSettings = namedtuple("RepulseSettings", ["enable_repulse", "debounce_ms"])


class DriverPlatform(BasePlatform, metaclass=abc.ABCMeta):

    """Baseclass for platforms with drivers."""

    __slots__ = []  # type: List[str]

    def __init__(self, machine):
        """Add driver feature and default max_pulse length."""
        super().__init__(machine)

        # Set default platform features. Each platform interface can change
        # these to notify the framework of the specific features it supports.
        self.features['has_drivers'] = True
        self.features['max_pulse'] = 255

    @abc.abstractmethod
    def configure_driver(self, config: DriverConfig, number: str, platform_settings: dict) -> "DriverPlatformInterface":
        """Subclass this method in a platform module to configure a driver.

        This method should return a reference to the driver's platform interface
        object which will be called to access the hardware.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def clear_hw_rule(self, switch: SwitchSettings, coil: DriverSettings):
        """Subclass this method in a platform module to clear a hardware switch rule for this switch.

        Clearing a hardware rule means actions on this switch will no longer
        affect coils.

        Another way to think of this is that it 'disables' a hardware rule.
        This is what you'd use to disable flippers and autofire_coils during
        tilt, game over, etc.

        """
        raise NotImplementedError

    @classmethod
    def get_coil_config_section(cls) -> Optional[str]:
        """Return addition config section for coils."""
        return None

    def validate_coil_section(self, driver, config) -> dict:
        """Validate coil config for platform."""
        if self.get_coil_config_section():
            spec = self.get_coil_config_section()   # pylint: disable-msg=assignment-from-none
            config = driver.machine.config_validator.validate_config(spec, config, driver.name)
        elif config:
            raise AssertionError("No platform_config supported but not empty {} for driver {}".
                                 format(config, driver.name))

        return config

    @abc.abstractmethod
    def set_pulse_on_hit_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit rule on driver.

        Pulses a driver when a switch is hit. When the switch is released the pulse continues. Typically used for
        autofire coils such as pop bumpers.
        """
        raise NotImplementedError

    # pylint: disable-msg=no-self-use
    def set_delayed_pulse_on_hit_rule(self, enable_switch: SwitchSettings, coil: DriverSettings, delay_ms: int):
        """Set pulse on hit and release rule to driver.

        When a switch is hit and a certain delay passed it pulses a driver.
        When the switch is released the pulse continues.
        Typically used for kickbacks.
        """
        del enable_switch
        del coil
        del delay_ms
        raise AssertionError("This platform does not support delayed pulse hardware rules.")

    @abc.abstractmethod
    def set_pulse_on_hit_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit and release rule to driver.

        Pulses a driver when a switch is hit. When the switch is released the pulse is canceled. Typically used on
        the main coil for dual coil flippers without eos switch.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def set_pulse_on_hit_and_enable_and_release_rule(self, enable_switch: SwitchSettings, coil: DriverSettings):
        """Set pulse on hit and enable and release rule on driver.

        Pulses a driver when a switch is hit. Then enables the driver (may be with pwm). When the switch is released
        the pulse is canceled and the driver gets disabled. Typically used for single coil flippers.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def set_pulse_on_hit_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                      eos_switch: SwitchSettings, coil: DriverSettings,
                                                      repulse_settings: Optional[RepulseSettings]):
        """Set pulse on hit and enable and release and disable rule on driver.

        Pulses a driver when a switch is hit. When the switch is released
        the pulse is canceled and the driver gets disabled. When the eos_switch is hit the pulse is canceled
        and the driver becomes disabled. Typically used on the main coil for dual-wound coil flippers with eos switch.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def set_pulse_on_hit_and_enable_and_release_and_disable_rule(self, enable_switch: SwitchSettings,
                                                                 eos_switch: SwitchSettings, coil: DriverSettings,
                                                                 repulse_settings: Optional[RepulseSettings]):
        """Set pulse on hit and enable and release and disable rule on driver.

        Pulses a driver when a switch is hit. Then enables the driver (may be with pwm). When the switch is released
        the pulse is canceled and the driver becomes disabled. When the eos_switch is hit the pulse is canceled
        and the driver becomes enabled (likely with PWM).
        Typically used on the coil for single-wound coil flippers with eos switch.
        """
        raise NotImplementedError



"""Device Mixins."""
from typing import List


class DevicePositionMixin():

    """Adds x/y/z getters to a device.

    For devices that have x/y/z config in the yaml this
    mixin will add getters to allow the use of device.x
    device.y and device.z instead of device.config['x']
    """

    __slots__ = []  # type: List[str]

    @property
    def x(self):
        """Get the X value from the config.

        Returns the devices x position from config
        """
        return self.config.get('x', None)

    @property
    def y(self):
        """Get the Y value from the config.

        Returns the devices y position from config
        """
        return self.config.get('y', None)

    @property
    def z(self):
        """Get the Z value from the config.

        Returns the devices z position from config
        """
        return self.config.get('z', None)



"""Test the bonus mode."""
from mpf.tests.MpfTestCase import MpfTestCase, MagicMock, test_config_directory


class TestBonusMode(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/bonus/'

    def _start_game(self):
        self.machine.playfield.add_ball = MagicMock()
        self.machine.ball_controller.num_balls_known = 3
        self.hit_and_release_switch("s_start")
        self.advance_time_and_run()
        self.assertIsNotNone(self.machine.game)

    def _stop_game(self):
        # stop game
        self.assertIsNotNone(self.machine.game)
        self.machine.game.end_game()
        self.advance_time_and_run()
        self.assertIsNone(self.machine.game)

    def testBonus(self):
        self.mock_event("bonus_ramps")
        self.mock_event("bonus_modes")
        self.mock_event("bonus_subtotal")
        self.mock_event("bonus_multiplier")
        self.mock_event("bonus_total")
        # start game
        self._start_game()

        self.post_event("start_mode1")
        self.advance_time_and_run()
        self.assertTrue(self.machine.mode_controller.is_active('mode1'))

        # player gets some score
        self.post_event("hit_target")

        # score 3 ramp shots
        self.post_event("score_ramps")
        self.post_event("score_ramps")
        self.post_event("score_ramps")

        # score 2 modes
        self.post_event("score_modes")
        self.post_event("score_modes")

        # increase multiplier to 5 (by hitting it 4 times)
        self.post_event("add_multiplier")
        self.post_event("add_multiplier")
        self.post_event("add_multiplier")
        self.post_event("add_multiplier")
        self.advance_time_and_run()

        # drain a ball
        self.machine.game.balls_in_play = 0
        self.advance_time_and_run(1)

        # check that bonus mode is loaded
        self.assertModeRunning('bonus')
        self.advance_time_and_run(29)
        self.assertEqual(3000, self._last_event_kwargs["bonus_ramps"]["score"])
        self.assertEqual(3, self._last_event_kwargs["bonus_ramps"]["hits"])
        self.assertEqual(10000, self._last_event_kwargs["bonus_modes"]["score"])
        self.assertEqual(2, self._last_event_kwargs["bonus_modes"]["hits"])
        self.assertEqual(13000, self._last_event_kwargs["bonus_subtotal"]["score"])
        self.assertEqual(5, self._last_event_kwargs["bonus_multiplier"]["multiplier"])
        self.assertEqual(65000, self._last_event_kwargs["bonus_total"]["score"])
        self.assertEqual(66337, self.machine.game.player.score)

        # check resets
        self.assertEqual(0, self.machine.game.player.ramps)
        self.assertEqual(2, self.machine.game.player.modes)
        self.assertEqual(5, self.machine.game.player.bonus_multiplier)

        self.mock_event("bonus_ramps")
        self.mock_event("bonus_modes")
        self.mock_event("bonus_subtotal")
        self.mock_event("bonus_multiplier")
        self.mock_event("bonus_total")

        # drain a ball
        self.machine.game.balls_in_play = 0
        self.advance_time_and_run(30)

        self.assertEqual(0, self._last_event_kwargs["bonus_ramps"]["score"])
        self.assertEqual(0, self._last_event_kwargs["bonus_ramps"]["hits"])
        self.assertEqual(10000, self._last_event_kwargs["bonus_modes"]["score"])
        self.assertEqual(2, self._last_event_kwargs["bonus_modes"]["hits"])
        self.assertEqual(10000, self._last_event_kwargs["bonus_subtotal"]["score"])
        self.assertEqual(5, self._last_event_kwargs["bonus_multiplier"]["multiplier"])
        self.assertEqual(50000, self._last_event_kwargs["bonus_total"]["score"])
        self.assertEqual(116337, self.machine.game.player.score)

        # multiplier should stay the same
        self.assertEqual(0, self.machine.game.player.ramps)
        self.assertEqual(2, self.machine.game.player.modes)
        self.assertEqual(5, self.machine.game.player.bonus_multiplier)

        # drain a ball
        self.machine.game.balls_in_play = 0
        self.advance_time_and_run(30)

        # this entry event should be in bonus even though it is 0
        self.assertIn('bonus_ramps', self._last_event_kwargs)
        self.assertIn('bonus_modes', self._last_event_kwargs)

        # make some changes for the next test
        self.machine.game.player.ramps = 1
        self.machine.game.player.bonus_multiplier = 2

        # test the hurry up
        self.mock_event("bonus_start")
        self.mock_event("bonus_ramps")
        self.mock_event("bonus_modes")
        self.mock_event("bonus_subtotal")
        self.mock_event("bonus_multiplier")
        self.mock_event("bonus_total")

        # drain a ball
        self.machine.game.balls_in_play = 0
        self.advance_time_and_run(1)

        self.assertEventCalled('bonus_start')
        self.assertEventNotCalled('bonus_ramps')
        self.assertEventNotCalled('bonus_modes')
        self.assertEventNotCalled('bonus_subtotal')
        self.assertEventNotCalled('bonus_multiplier')
        self.assertEventNotCalled('bonus_total')

        self.advance_time_and_run(2)

        self.assertEventCalled('bonus_ramps')
        self.assertEventNotCalled('bonus_modes')
        self.assertEventNotCalled('bonus_subtotal')
        self.assertEventNotCalled('bonus_multiplier')
        self.assertEventNotCalled('bonus_total')

        self.post_event('flipper_cancel', .1)
        self.assertEventCalled('bonus_modes')
        self.assertEventNotCalled('bonus_subtotal')
        self.assertEventNotCalled('bonus_multiplier')
        self.assertEventNotCalled('bonus_total')

        self.advance_time_and_run(.5)
        self.assertEventCalled('bonus_subtotal')
        self.assertEventNotCalled('bonus_multiplier')
        self.assertEventNotCalled('bonus_total')

        self.advance_time_and_run(.5)
        self.assertEventCalled('bonus_multiplier')
        self.assertEventNotCalled('bonus_total')

        self.advance_time_and_run(.5)
        self.assertEventCalled('bonus_total')

        # test multiplier screens are skipped if multiplier is 1
        self.advance_time_and_run(30)
        self.machine.game.player.bonus_multiplier = 1
        self.machine.game.player.ramps = 1

        # test the hurry up
        self.mock_event("bonus_start")
        self.mock_event("bonus_ramps")
        self.mock_event("bonus_modes")
        self.mock_event("bonus_subtotal")
        self.mock_event("bonus_multiplier")
        self.mock_event("bonus_total")

        # drain a ball
        self.machine.game.balls_in_play = 0
        self.advance_time_and_run(30)

        self.assertEventCalled('bonus_start')
        self.assertEventCalled('bonus_ramps')
        self.assertEventCalled('bonus_modes')
        self.assertEventNotCalled('bonus_subtotal')
        self.assertEventNotCalled('bonus_multiplier')
        self.assertEventCalled('bonus_total')

    @test_config_directory("tests/machine_files/bonus_no_keep_multiplier/")
    def testBonusNoKeepMultiplier(self):
        self.mock_event("bonus_ramps")
        self.mock_event("bonus_modes")
        self.mock_event("bonus_subtotal")
        self.mock_event("bonus_multiplier")
        self.mock_event("bonus_total")
        # start game
        self._start_game()

        self.post_event("start_mode1")
        self.advance_time_and_run()
        self.assertTrue(self.machine.mode_controller.is_active('mode1'))

        # player gets some score
        self.post_event("hit_target")

        # score 3 ramp shots
        self.post_event("score_ramps")
        self.post_event("score_ramps")
        self.post_event("score_ramps")

        # score 2 modes
        self.post_event("score_modes")
        self.post_event("score_modes")

        # increase multiplier to 5 (by hitting it 4 times)
        self.post_event("add_multiplier")
        self.post_event("add_multiplier")
        self.post_event("add_multiplier")
        self.post_event("add_multiplier")
        self.advance_time_and_run()

        # drain a ball
        self.machine.game.balls_in_play = 0
        self.advance_time_and_run(1)

        # check that bonus mode is loaded
        self.assertModeRunning('bonus')
        self.advance_time_and_run(29)
        self.assertEqual(3000, self._last_event_kwargs["bonus_ramps"]["score"])
        self.assertEqual(3, self._last_event_kwargs["bonus_ramps"]["hits"])
        self.assertEqual(10000, self._last_event_kwargs["bonus_modes"]["score"])
        self.assertEqual(2, self._last_event_kwargs["bonus_modes"]["hits"])
        self.assertEqual(13000, self._last_event_kwargs["bonus_subtotal"]["score"])
        self.assertEqual(5, self._last_event_kwargs["bonus_multiplier"]["multiplier"])
        self.assertEqual(65000, self._last_event_kwargs["bonus_total"]["score"])
        self.assertEqual(66337, self.machine.game.player.score)

        # check resets
        self.assertEqual(0, self.machine.game.player.ramps)
        self.assertEqual(2, self.machine.game.player.modes)
        self.assertEqual(1, self.machine.game.player.bonus_multiplier)

        self.mock_event("bonus_ramps")
        self.mock_event("bonus_modes")
        self.mock_event("bonus_subtotal")
        self.mock_event("bonus_multiplier")
        self.mock_event("bonus_total")

        # drain a ball
        self.machine.game.balls_in_play = 0
        self.advance_time_and_run(30)

        self.assertNotIn('bonus_ramps', self._last_event_kwargs)
        self.assertEqual(10000, self._last_event_kwargs["bonus_modes"]["score"])
        self.assertEqual(2, self._last_event_kwargs["bonus_modes"]["hits"])
        self.assertNotIn('bonus_subtotal', self._last_event_kwargs)
        self.assertNotIn('bonus_multiplier', self._last_event_kwargs)
        self.assertEqual(10000, self._last_event_kwargs["bonus_total"]["score"])
        self.assertEqual(76337, self.machine.game.player.score)


    def testBonusTilted(self):
        self.mock_event("bonus_ramps")
        self.mock_event("bonus_modes")
        self.mock_event("bonus_subtotal")
        self.mock_event("bonus_multiplier")
        self.mock_event("bonus_total")
        # start game
        self._start_game()

        self.post_event("start_mode1")
        self.advance_time_and_run()

        # player gets some score
        self.post_event("hit_target")

        # score 3 ramp shots
        self.post_event("score_ramps")
        self.post_event("score_ramps")
        self.post_event("score_ramps")

        # score 2 modes
        self.post_event("score_modes")
        self.post_event("score_modes")

        # increase multiplier to 5 (by hitting it 4 times)
        self.post_event("add_multiplier")
        self.post_event("add_multiplier")
        self.post_event("add_multiplier")
        self.post_event("add_multiplier")
        self.advance_time_and_run()

        # tilt the game
        self.machine.game.tilted = True

        # drain a ball
        self.machine.game.balls_in_play = 0
        self.advance_time_and_run(30)
        self.assertEqual(1337, self.machine.game.player.score)

        # check resets
        self.assertEqual(0, self.machine.game.player.ramps)
        self.assertEqual(2, self.machine.game.player.modes)
        self.assertEqual(5, self.machine.game.player.bonus_multiplier)

    @test_config_directory("tests/machine_files/bonus_additional_events/")
    def testBonusStopEvent(self):
        self.mock_event("bonus_ramps")
        self.mock_event("bonus_modes")
        self.mock_event("bonus_subtotal")
        self.mock_event("bonus_multiplier")
        self.mock_event("bonus_total")
        # start game
        self._start_game()

        self.post_event("start_mode1")
        self.advance_time_and_run()

        # player gets some score
        self.post_event("hit_target")

        # score 3 ramp shots
        self.post_event("score_ramps")
        self.post_event("score_ramps")
        self.post_event("score_ramps")

        # score 2 modes
        self.post_event("score_modes")
        self.post_event("score_modes")

        # increase multiplier to 5 (by hitting it 4 times)
        self.post_event("add_multiplier")
        self.post_event("add_multiplier")
        self.post_event("add_multiplier")
        self.post_event("add_multiplier")
        self.advance_time_and_run()

        # drain a ball
        self.machine.game.balls_in_play = 0
        self.advance_time_and_run(30)
        self.assertEqual(66337, self.machine.game.player.score)
        self.assertModeRunning("bonus")
        self.post_event("stop_bonus")
        self.advance_time_and_run(.1)
        self.assertModeNotRunning("bonus")

        # check resets
        self.assertEqual(0, self.machine.game.player.ramps)
        self.assertEqual(2, self.machine.game.player.modes)
        self.assertEqual(5, self.machine.game.player.bonus_multiplier)

    @test_config_directory("tests/machine_files/bonus_dynamic_keep_multiplier/")
    def testBonusDynamicKeepMultiplier(self):
        # start game
        self._start_game()

        self.post_event("start_mode1")
        self.advance_time_and_run()
        self.assertTrue(self.machine.mode_controller.is_active('mode1'))

        # increase multiplier to 5 (by hitting it 4 times)
        self.post_event("add_multiplier")
        self.post_event("add_multiplier")
        self.post_event("add_multiplier")
        self.post_event("add_multiplier")
        self.advance_time_and_run()

        # drain a ball
        self.machine.game.balls_in_play = 0
        self.advance_time_and_run(1)

        # check that bonus mode is loaded
        self.assertModeRunning('bonus')
        self.advance_time_and_run(29)

        # check resets
        self.assertEqual(5, self.machine.game.player.bonus_multiplier)

        # drain a ball
        self.machine.game.balls_in_play = 0
        self.advance_time_and_run(1)

        # check that bonus mode is loaded
        self.assertModeRunning('bonus')
        self.advance_time_and_run(29)

        # add multipliers
        self.post_event("add_multiplier")
        self.post_event("add_multiplier")
        self.post_event("add_multiplier")
        self.post_event("add_multiplier")
        self.advance_time_and_run()

        # check resets
        self.assertEqual(1, self.machine.game.player.bonus_multiplier)




import inspect

from mpf.core.utility_functions import Util
from mpf.tests.MpfTestCase import MpfTestCase


class TestDeviceManager(MpfTestCase):

    def test_control_events_arguments(self):
        for device_type in self.machine.config['mpf']['device_modules']:

            device_cls = Util.string_to_class(device_type)

            config_spec = self.machine.config_validator.config_spec[device_cls.config_section]

            for k in config_spec:
                if not k.endswith('_events') or k == "control_events" or config_spec[k] == "ignore":
                    continue
                method_name = k[:-7]
                method = getattr(device_cls, "event_{}".format(method_name), None)
                self.assertIsNotNone(method, "Method {}.event_{} is missing for {}".format(device_type, method_name, k))

                sig = inspect.signature(method)

                self.assertTrue(sig.parameters['self'],
                    "Method {}.{} is missing self. Actual signature: {}".format(
                    device_type, method_name, sig))

                self.assertTrue('kwargs' in sig.parameters,
                    "Method {}.{} is missing **kwargs. Actual signature: {}".format(
                    device_type, method_name, sig))

                self.assertEqual(sig.parameters['kwargs'].kind, inspect._VAR_KEYWORD,
                    "Method {}.{} kwargs param is missing '**'".format(
                    device_type, method_name))

                self.assertTrue(hasattr(method, "relative_priority"),
                                "Method {}.{} is missing a relative_priority. Did you apply the event_handler "
                                "decorator?".format(device_type, method_name))


"""Device that implements a ball save."""
from typing import Optional

from mpf.core.delays import DelayManager
from mpf.core.device_monitor import DeviceMonitor
from mpf.core.events import event_handler
from mpf.core.mode import Mode
from mpf.core.mode_device import ModeDevice
from mpf.core.system_wide_device import SystemWideDevice

MYPY = False
if MYPY:   # pragma: no cover
    from mpf.core.machine import MachineController  # pylint: disable-msg=cyclic-import,unused-import
    from mpf.devices.playfield import Playfield     # pylint: disable-msg=cyclic-import,unused-import


@DeviceMonitor("saves_remaining", "enabled", "timer_started", "state")
class BallSave(SystemWideDevice, ModeDevice):

    """Ball save device which will give back the ball within a certain time."""

    config_section = 'ball_saves'
    collection = 'ball_saves'
    class_label = 'ball_save'

    __slots__ = ["active_time", "unlimited_saves", "source_playfield", "delay", "enabled", "timer_started",
                 "saves_remaining", "early_saved", "state", "_scheduled_balls"]

    def __init__(self, machine: "MachineController", name: str) -> None:
        """Initialise ball save."""
        self.unlimited_saves = None         # type: Optional[bool]
        self.source_playfield = None        # type: Optional[Playfield]
        super().__init__(machine, name)

        self.delay = DelayManager(machine)
        self.enabled = False
        self.timer_started = False
        self.saves_remaining = 0
        self.early_saved = 0
        self.state = 'disabled'
        self._scheduled_balls = 0
        self.active_time = 0

    async def _initialize(self) -> None:
        await super()._initialize()
        self.unlimited_saves = self.config['balls_to_save'] == -1
        self.source_playfield = self.config['source_playfield']

    @property
    def can_exist_outside_of_game(self) -> bool:
        """Return true if this device can exist outside of a game."""
        return True

    def validate_and_parse_config(self, config: dict, is_mode_config: bool, debug_prefix: str = None) -> dict:
        """Make sure timer_start_events are not in enable_events."""
        config = super().validate_and_parse_config(config, is_mode_config, debug_prefix)

        for event in config['timer_start_events']:
            if event in config['enable_events']:
                raise AssertionError("{}: event {} in timer_start_events will not work because it is also in "
                                     "enable_events. Omit it!".format(event, str(self)))

        if config['delayed_eject_events'] and config['eject_delay']:
            raise AssertionError("cannot use delayed_eject_events and eject_delay at the same time.")

        return config

    def enable(self) -> None:
        """Enable ball save."""
        super().enable()
        if self.enabled:
            return

        self.saves_remaining = self.config['balls_to_save']
        self.early_saved = 0
        self.enabled = True
        self.state = 'enabled'
        self.active_time = self.config['active_time'].evaluate([])
        self.debug_log("Enabling. Auto launch: {}, Balls to save: {}, Active time: {}s".format(
                       self.config['auto_launch'],
                       self.config['balls_to_save'],
                       self.active_time))

        # Enable shoot again
        self.machine.events.add_handler('ball_drain',
                                        self._ball_drain_while_active,
                                        priority=1000)

        if (self.active_time > 0 and
                not self.config['timer_start_events']):
            self.timer_start()

        self.machine.events.post('ball_save_{}_enabled'.format(self.name))
        '''event: ball_save_(name)_enabled
        desc: The ball save called (name) has just been enabled.
        '''

    @event_handler(1)
    def event_disable(self, **kwargs):
        """Event handler for disable event."""
        del kwargs
        self.disable()

    def disable(self) -> None:
        """Disable ball save."""
        if not self.enabled:
            return

        self.enabled = False
        self.state = 'disabled'
        self.timer_started = False
        self.debug_log("Disabling...")
        self.machine.events.remove_handler(self._ball_drain_while_active)
        self.delay.remove('disable')
        self.delay.remove('hurry_up')
        self.delay.remove('grace_period')

        self.machine.events.post('ball_save_{}_disabled'.format(self.name))
        '''event: ball_save_(name)_disabled
        desc: The ball save called (name) has just been disabled.
        '''

    @event_handler(9)
    def event_timer_start(self, **kwargs):
        """Event handler for timer start event."""
        del kwargs
        self.timer_start()

    def timer_start(self) -> None:
        """Start the timer.

        This is usually called after the ball was ejected while the ball save may have been enabled earlier.
        """
        if self.timer_started or not self.enabled:
            return

        self.timer_started = True

        self.machine.events.post('ball_save_{}_timer_start'.format(self.name))
        '''event: ball_save_(name)_timer_start
        desc: The ball save called (name) has just start its countdown timer.
        '''

        if self.active_time > 0:
            self.debug_log('Starting ball save timer: %ss',
                           self.active_time)
            active_time_ms = self.active_time * 1000
            self.delay.add(name='disable',
                           ms=(active_time_ms +
                               self.config['grace_period']),
                           callback=self.disable)
            self.delay.add(name='grace_period',
                           ms=active_time_ms,
                           callback=self._grace_period)
            self.delay.add(name='hurry_up',
                           ms=(active_time_ms -
                               self.config['hurry_up_time']),
                           callback=self._hurry_up)

    def _hurry_up(self) -> None:
        self.debug_log("Starting Hurry Up")

        self.state = 'hurry_up'

        self.machine.events.post('ball_save_{}_hurry_up'.format(self.name))
        '''event: ball_save_(name)_hurry_up
        desc: The ball save called (name) has just entered its hurry up mode.
        '''

    def _grace_period(self) -> None:
        self.debug_log("Starting Grace Period")

        self.state = 'grace_period'

        self.machine.events.post('ball_save_{}_grace_period'.format(self.name))
        '''event: ball_save_(name)_grace_period
        desc: The ball save called (name) has just entered its grace period
            time.
        '''

    def _get_number_of_balls_to_save(self, available_balls: int) -> int:
        if self.machine.game and self.machine.game.balls_in_play > 0:
            if self.config['only_last_ball'] and self.machine.game.balls_in_play > 1:
                self.debug_log("Will only save last ball but %s are in play.", self.machine.game.balls_in_play)
                return 0
        else:
            self.debug_log("Received request to save ball, but no balls are in"
                           " play. Discarding request.")
            return 0

        balls_to_save = available_balls

        if self.config['only_last_ball'] and balls_to_save > 1:
            balls_to_save = 1

        if balls_to_save > self.machine.game.balls_in_play:
            balls_to_save = self.machine.game.balls_in_play

        if balls_to_save > self.saves_remaining and not self.unlimited_saves:
            balls_to_save = self.saves_remaining

        return balls_to_save

    def _reduce_remaining_saves_and_disable_if_zero(self, balls_to_save: int) -> None:
        if not self.unlimited_saves:
            self.saves_remaining -= balls_to_save
            self.debug_log("Saves remaining: %s", self.saves_remaining)
        else:
            self.debug_log("Unlimited saves remaining")

        if self.saves_remaining <= 0 and not self.unlimited_saves:
            self.debug_log("Disabling since there are no saves remaining")
            self.disable()

    def _ball_drain_while_active(self, balls: int, **kwargs) -> Optional[dict]:
        del kwargs
        if balls <= 0:
            return {}

        balls_to_save = self._get_number_of_balls_to_save(balls)

        self.debug_log("Ball(s) drained while active. Requesting new one(s). "
                       "Auto launch: %s", self.config['auto_launch'])

        self.machine.events.post('ball_save_{}_saving_ball'.format(self.name),
                                 balls=balls_to_save, early_save=False)
        '''event: ball_save_(name)_saving_ball
        desc: The ball save called (name) has just saved one (or more) balls.

        args:
            balls: The number of balls this ball saver is saving.
            early_save: True if this is an early ball save.
        '''

        self._schedule_balls(balls_to_save)

        self._reduce_remaining_saves_and_disable_if_zero(balls_to_save)

        return {'balls': balls - balls_to_save}

    @event_handler(8)
    def event_early_ball_save(self, **kwargs):
        """Event handler for early_ball_save event."""
        del kwargs
        self.early_ball_save()

    def early_ball_save(self) -> None:
        """Perform early ball save if enabled."""
        if not self.enabled:
            return

        if not self._get_number_of_balls_to_save(1):
            return

        if self.early_saved > 0:
            self.debug_log("Already performed an early ball save. Ball needs to drain first.")
            return

        self.machine.events.post('ball_save_{}_saving_ball'.format(self.name),
                                 balls=1, early_save=True)
        # doc block above

        self.debug_log("Performing early ball save.")
        self.early_saved += 1
        self._schedule_balls(1)
        self.machine.events.add_handler('ball_drain',
                                        self._early_ball_save_drain_handler,
                                        priority=1001)

        self._reduce_remaining_saves_and_disable_if_zero(1)

    def _early_ball_save_drain_handler(self, balls: int, **kwargs) -> dict:
        del kwargs
        if self.early_saved and balls > 0:
            balls -= 1
            self.early_saved -= 1
            self.debug_log("Early saved ball drained.")
            self.machine.events.remove_handler(self._early_ball_save_drain_handler)
            return {'balls': balls}

        return {}

    def _schedule_balls(self, balls_to_save: int) -> None:
        if self.config['eject_delay']:
            # schedule after delay. to add some drama
            self.delay.add(self.config['eject_delay'], self._add_balls, balls_to_save=balls_to_save)
        elif self.config['delayed_eject_events']:
            # unlimited delay. wait for event
            self._scheduled_balls += balls_to_save
        else:
            # default: no delay. just eject balls right now
            self._add_balls(balls_to_save)

    @event_handler(4)
    def event_delayed_eject(self, **kwargs):
        """Event handler for delayed_eject event."""
        del kwargs
        self.delayed_eject()

    def delayed_eject(self):
        """Trigger eject of all scheduled balls."""
        self._add_balls(self._scheduled_balls)
        self._scheduled_balls = 0

    def _add_balls(self, balls_to_save, **kwargs):
        del kwargs
        self.source_playfield.add_ball(balls=balls_to_save,
                                       player_controlled=self.config['auto_launch'] ^ 1)

    def device_removed_from_mode(self, mode: Mode) -> None:
        """Disable ball save when mode ends."""
        super().device_removed_from_mode(mode)
        self.debug_log("Removing...")

        self.disable()

        if self.config['delayed_eject_events']:
            self.debug_log("Triggering delayed eject because mode ended.")
            self.delayed_eject()



import asyncio
from typing import Tuple

from mpf.core.bcp.bcp_socket_client import decode_command_string, encode_command_string

from mpf.tests.loop import MockServer, MockQueueSocket

from mpf.tests.MpfTestCase import MpfTestCase


class TestVirtualPinball(MpfTestCase):

    def __init__(self, methodName):
        super().__init__(methodName)
        # remove config patch which disables bcp
        del self.machine_config_patches['bcp']
        self.machine_config_patches['bcp'] = dict()
        self.machine_config_patches['bcp']['connections'] = []

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/virtual_pinball/'

    def get_use_bcp(self):
        return True

    def _mock_loop(self):
        self.mock_server = MockServer(self.clock.loop)
        self.clock.mock_server("127.0.0.1", 5051, self.mock_server)

    @staticmethod
    async def _get_and_decode(client) -> Tuple[str, dict]:
        data = await client.send_queue.get()
        return decode_command_string(data[0:-1].decode())

    def _encode_and_send(self, client, cmd, **kwargs):
        client.recv_queue.append((encode_command_string(cmd, **kwargs) + '\n').encode())

    def test_virtual_pinball(self):
        # connect a client
        client = MockQueueSocket(self.loop)
        self.loop.run_until_complete(self.mock_server.add_client(client))

        # check hello
        cmd, args = self.loop.run_until_complete(self._get_and_decode(client))
        self.assertEqual("hello", cmd)

        self._encode_and_send(client, "monitor_start", category="switches")
        self._encode_and_send(client, "monitor_start", category="drivers")
        self._encode_and_send(client, "monitor_start", category="devices")

        self.advance_time_and_run()
        client.send_queue = asyncio.Queue()

        self.machine.lights["test_light1"].on()

        cmd, args = self.loop.run_until_complete(self._get_and_decode(client))
        self.assertEqual("device", cmd)
        self.assertEqual("test_light1", args['name'])
        self.assertEqual("light", args['type'])
        self.assertEqual({'color': [255, 255, 255]}, args['state'])

        self.machine.coils["c_test"].pulse()
        self.advance_time_and_run()

        cmd, args = self.loop.run_until_complete(self._get_and_decode(client))
        self.assertEqual("driver_event", cmd)
        self.assertEqual("pulse", args['action'])
        self.assertEqual("c_test", args['name'])
        self.assertEqual("0-0", args['number'])
        self.assertEqual(23, args['pulse_ms'])
        self.assertEqual(1.0, args['pulse_power'])

        self.machine.coils["c_test_allow_enable"].enable()
        cmd, args = self.loop.run_until_complete(self._get_and_decode(client))
        self.assertEqual("driver_event", cmd)
        self.assertEqual("enable", args['action'])
        self.assertEqual("c_test_allow_enable", args['name'])
        self.assertEqual("0-1", args['number'])
        self.assertEqual(23, args['pulse_ms'])
        self.assertEqual(1.0, args['pulse_power'])
        self.assertEqual(1.0, args['hold_power'])

        self.machine.coils["c_test_allow_enable"].disable()
        cmd, args = self.loop.run_until_complete(self._get_and_decode(client))
        self.assertEqual("driver_event", cmd)
        self.assertEqual("disable", args['action'])
        self.assertEqual("c_test_allow_enable", args['name'])
        self.assertEqual("0-1", args['number'])

        self.machine.flippers["f_test_single"].enable()
        cmd, args = self.loop.run_until_complete(self._get_and_decode(client))
        self.assertEqual("device", cmd)
        self.assertEqual("f_test_single", args['name'])
        self.assertEqual("flipper", args['type'])
        self.assertEqual({"enabled": True}, args['state'])

        cmd, args = self.loop.run_until_complete(self._get_and_decode(client))
        self.assertEqual("driver_event", cmd)
        self.assertEqual({'enable_switch_invert': False,
                          'coil_name': 'c_flipper_main',
                          'coil_pulse_power': 1.0,
                          'coil_hold_power': 'HoldSettings(power=0.375)',
                          'enable_switch_name': 's_flipper',
                          'enable_switch_number': '0-3',
                          'coil_pulse_ms': 10,
                          'coil_number': '0-3',
                          'action': 'pulse_on_hit_and_enable_and_release',
                          'coil_recycle': False,
                          'enable_switch_debounce': False}, args)

        self.machine.flippers["f_test_single"].disable()
        cmd, args = self.loop.run_until_complete(self._get_and_decode(client))
        self.assertEqual("driver_event", cmd)
        self.assertEqual({'coil_number': '0-3',
                          'enable_switch_invert': False,
                          'action': 'remove',
                          'enable_switch_number': '0-3'}, args)

        cmd, args = self.loop.run_until_complete(self._get_and_decode(client))
        self.assertEqual("device", cmd)
        self.assertEqual("f_test_single", args['name'])
        self.assertEqual("flipper", args['type'])
        self.assertEqual({"enabled": False}, args['state'])

        self.assertSwitchState("s_test_nc", 0)

        self._encode_and_send(client, "switch", name="s_test_nc", state=1)
        self.advance_time_and_run()
        self.assertSwitchState("s_test_nc", 1)



"""Contains the FileManager base classes."""

import logging
import os

from mpf.file_interfaces.pickle_interface import PickleInterface
from mpf.file_interfaces.yaml_interface import YamlInterface

MYPY = False
if MYPY:    # pragma: no cover
    from typing import Dict, List  # pylint: disable-msg=cyclic-import,unused-import


class FileManager:

    """Manages file interfaces."""

    __slots__ = []  # type: List[str]

    log = logging.getLogger('FileManager')
    file_interfaces = dict()    # type: Dict[str, YamlInterface]
    initialized = False

    @classmethod
    def init(cls):
        """Initialise file interfaces."""
        cls.file_interfaces[".yaml"] = YamlInterface()
        cls.file_interfaces[".bin"] = PickleInterface()

        FileManager.initialized = True

    @staticmethod
    def locate_file(filename) -> str:
        """Find a file location.

        Args:
        ----
            filename: Filename to locate

        Returns: Location of file
        """
        if not filename:
            raise FileNotFoundError("No filename provided")

        if not FileManager.initialized:
            FileManager.init()

        ext = os.path.splitext(filename)[1]

        if not os.path.isfile(filename):
            # If the file doesn't have an extension, let's see if we can find
            # one
            if not ext:
                for config_processor in set(FileManager.file_interfaces.values()):
                    questionable_file, ext = config_processor.find_file(filename)
                    if isinstance(questionable_file, str):
                        return questionable_file

            raise FileNotFoundError("File not found: {}".format(filename))

        return filename

    @staticmethod
    def get_file_interface(filename):
        """Return a file interface."""
        try:
            FileManager.locate_file(filename)
        except FileNotFoundError:
            return None

        ext = os.path.splitext(filename)[1]

        try:
            return FileManager.file_interfaces[ext]
        except KeyError:
            return None

    @staticmethod
    def load(filename, verify_version=False, halt_on_error=True):
        """Load a file by name."""
        if not FileManager.initialized:
            FileManager.init()

        try:
            file = FileManager.locate_file(filename)
        except FileNotFoundError:
            if halt_on_error:
                raise IOError("Could not find file {}".format(filename))

            return dict()

        if not file and halt_on_error:
            raise IOError(
                "Could not find file '{}'. Resolved abs path to {}".format(
                    filename, os.path.abspath(filename)))

        ext = os.path.splitext(file)[1]

        try:
            interface = FileManager.file_interfaces[ext]
        except KeyError:
            raise AssertionError("No config file processor available for file type {}".format(ext))

        return interface.load(file, verify_version, halt_on_error)

    @staticmethod
    def save(filename, data):
        """Save data to file."""
        if not FileManager.initialized:
            FileManager.init()

        ext = os.path.splitext(filename)[1]

        # save to temp file and move afterwards. prevents broken files
        temp_file = os.path.dirname(filename) + os.sep + "_" + os.path.basename(filename)

        try:
            FileManager.file_interfaces[ext].save(temp_file, data)
        except KeyError:
            raise AssertionError("No config file processor available for file type {}".format(ext))

        # move temp file
        os.replace(temp_file, filename)



from mpf.tests.MpfTestCase import MpfTestCase


class TestLightSegmentDisplays(MpfTestCase):

    def get_config_file(self):
        return 'config.yaml'

    def get_machine_path(self):
        return 'tests/machine_files/light_segment_displays/'

    def get_platform(self):
        return None

    def test_scoring(self):
        display1 = self.machine.segment_displays["display1"]

        # this should show the last two characters 37
        display1.add_text("1337")
        self.advance_time_and_run()

        self.assertLightColor("segment1_a", "on")
        self.assertLightColor("segment1_b", "on")
        self.assertLightColor("segment1_c", "on")
        self.assertLightColor("segment1_d", "on")
        self.assertLightColor("segment1_e", "off")
        self.assertLightColor("segment1_f", "off")
        self.assertLightColor("segment1_g", "on")
        self.assertLightColor("segment2_a", "on")
        self.assertLightColor("segment2_b", "on")
        self.assertLightColor("segment2_c", "on")
        self.assertLightColor("segment2_d", "off")
        self.assertLightColor("segment2_e", "off")
        self.assertLightColor("segment2_f", "off")
        self.assertLightColor("segment2_g", "off")

        # turn on all lights
        display1.add_text("88")
        self.advance_time_and_run()

        self.assertLightColor("segment1_a", "on")
        self.assertLightColor("segment1_b", "on")
        self.assertLightColor("segment1_c", "on")
        self.assertLightColor("segment1_d", "on")
        self.assertLightColor("segment1_e", "on")
        self.assertLightColor("segment1_f", "on")
        self.assertLightColor("segment1_g", "on")
        self.assertLightColor("segment2_a", "on")
        self.assertLightColor("segment2_b", "on")
        self.assertLightColor("segment2_c", "on")
        self.assertLightColor("segment2_d", "on")
        self.assertLightColor("segment2_e", "on")
        self.assertLightColor("segment2_f", "on")
        self.assertLightColor("segment2_g", "on")

        # back to four lights only
        display1.add_text("11")
        self.advance_time_and_run()

        self.assertLightColor("segment1_a", "off")
        self.assertLightColor("segment1_b", "on")
        self.assertLightColor("segment1_c", "on")
        self.assertLightColor("segment1_d", "off")
        self.assertLightColor("segment1_e", "off")
        self.assertLightColor("segment1_f", "off")
        self.assertLightColor("segment1_g", "off")
        self.assertLightColor("segment2_a", "off")
        self.assertLightColor("segment2_b", "on")
        self.assertLightColor("segment2_c", "on")
        self.assertLightColor("segment2_d", "off")
        self.assertLightColor("segment2_e", "off")
        self.assertLightColor("segment2_f", "off")
        self.assertLightColor("segment2_g", "off")

        # set invalid chars (for 7segment). should be "1 "
        display1.add_text("1{}".format(chr(244)))
        self.advance_time_and_run()

        self.assertLightColor("segment1_a", "off")
        self.assertLightColor("segment1_b", "on")
        self.assertLightColor("segment1_c", "on")
        self.assertLightColor("segment1_d", "off")
        self.assertLightColor("segment1_e", "off")
        self.assertLightColor("segment1_f", "off")
        self.assertLightColor("segment1_g", "off")
        self.assertLightColor("segment2_a", "off")
        self.assertLightColor("segment2_b", "off")
        self.assertLightColor("segment2_c", "off")
        self.assertLightColor("segment2_d", "off")
        self.assertLightColor("segment2_e", "off")
        self.assertLightColor("segment2_f", "off")
        self.assertLightColor("segment2_g", "off")

        display2 = self.machine.segment_displays["display2"]

        self.assertLightColor("segment3_x0", "off")
        self.assertLightColor("segment3_x1", "off")
        self.assertLightColor("segment3_x2", "off")
        self.assertLightColor("segment3_x3", "off")
        self.assertLightColor("segment4_x0", "off")
        self.assertLightColor("segment4_x1", "off")
        self.assertLightColor("segment4_x2", "off")
        self.assertLightColor("segment4_x3", "off")

        # this should translate 87 to bcd 0001 1110
        display2.add_text("87")
        self.advance_time_and_run()

        self.assertLightColor("segment3_x0", "off")
        self.assertLightColor("segment3_x1", "off")
        self.assertLightColor("segment3_x2", "off")
        self.assertLightColor("segment3_x3", "on")
        self.assertLightColor("segment4_x0", "on")
        self.assertLightColor("segment4_x1", "on")
        self.assertLightColor("segment4_x2", "on")
        self.assertLightColor("segment4_x3", "off")

        display2.set_flashing(True)
        self.assertLightFlashing("segment3_x3", "white", secs=2)



"""GI on fast."""
import logging

from mpf.core.utility_functions import Util
from mpf.platforms.interfaces.light_platform_interface import LightPlatformSoftwareFade


class FASTGIString(LightPlatformSoftwareFade):

    """A FAST GI string in a WPC machine."""

    __slots__ = ["log", "send"]

    def __init__(self, number, sender, machine, software_fade_ms: int) -> None:
        """Initialise GI string."""
        super().__init__(number, machine.clock.loop, software_fade_ms)
        self.log = logging.getLogger('FASTGIString.0x' + str(number))
        self.send = sender

    def set_brightness(self, brightness: float):
        """Set GI string to a certain brightness."""
        brightness = int(brightness * 255)
        if brightness >= 255:
            brightness = 255

        self.log.debug("Turning On GI String to brightness %s", brightness)
        # self.send('GI:' + self.number + ',' + Util.int_to_hex_string(brightness))

        self.send('GI:{},{}'.format(self.number,
                                    Util.int_to_hex_string(brightness)))

    def get_board_name(self):
        """Return the board of this light."""
        return "FAST WPC"

    def is_successor_of(self, other):
        """Return true if the other light has the previous number."""
        raise AssertionError("Not possible in FASTGI.")

    def get_successor_number(self):
        """Return next number."""
        raise AssertionError("Not possible in FASTGI.")

    def __lt__(self, other):
        """Order lights by string."""
        return self.number < other.number
